"""
This module implements the 7 constraints that are needed to train a
nacl.model.SALT2Like model on a dataset.

All these constraints are outlined and discussed in Guy's thesis (see. )

The contraints are implemented as quadratic penalties that are added to the
:math:`\chi^2`. These penalities are typically of the form:

.. math ::
    (f(\theta) - \alpha)^2

where :math:`f(\\theta)` is a function of the parameters, and :math:`\\alpha`
is a number.

Since there are several constraints, it is convenient to express them in vector
form:

.. math ::
    (\vec{F}(\\theta) - \vec{\alpha})^T \cdot (\vec{F}(\theta) - \vec{\alpha})


"""
import numpy as np
from scipy.sparse import coo_matrix, dok_matrix
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

class AllParametersFixedError(Exception): pass


def solve_constraints(cons, pars):
    """
    """
    H = cons.get_linearized_constraints(pars.free)
    rhs = cons.get_rhs()
    Q,R = np.linalg.qr(H.T.todense())
    dx = np.linalg.solve(R.T, rhs - H @ pars.full)
    dx = np.array(Q.dot(dx)).squeeze()

    pp = pars.copy()
    pp.full[:] = dx

    return pp.free


def ab_flux_at_10Mpc(Mb=-19.5):
    return 10**(-0.4 * (30+Mb))


class Constraint:
    """Generic contraint
    """
    def __init__(self, model, rhs):  # , pars_struct=None):
        """
        """
        self.model = model
        self.rhs = rhs
        # if pars_struct:
        #     self.pars_struct = pars_struct
        # else:
        #     self.pars_struct = self.model.pars.copy()
        #     self.pars_struct.release()

    def __call__(self, p=None, deriv=False):
        pass

class LinearConstraint(Constraint):
    """Generic linear contraints
    """
    def __init__(self, model, rhs):
        """
        """
        super().__init__(model, rhs)
        self.h_matrix = None
        self.rhs = rhs

    def init_h_matrix(self, pars):
        raise NotImplementedError()

    def init_pars(self, pars):
        """
        """
        self.h_matrix = self.init_h_matrix(pars)

    def __call__(self, pars, deriv=False):
        """evaluate the constraint
        """
        if self.h_matrix is None:
            self.h_matrix = self.init_h_matrix(pars)

        cons = self.h_matrix @ pars.full - self.rhs
        cons = float(cons)
        if not deriv:
            return cons
        return cons, self.h_matrix, None


class ConstraintSet:
    """Combine a series of constraints (linear or not)

    This class combines a set of constraints and produces a (quadratic)
    penality, added to the Log Likelihood. Compute the gradient and the hessian
    of this penality if required.
    """

    def __init__(self, constraints, mu=1.E10):
        """constructor
        """
        # self.model = model
        self.constraints = constraints
        self.mu = mu

    def init_pars(self, pars):
        """
        """
        for c in self.constraints:
            c.init_pars(pars)

    def __call__(self, pars, deriv=False):
        """evaluate the penality
        """
        npars = len(pars.full)

        pen = 0.
        # if no derivatives specified, return the sum of the quadratic
        # penalities associated with each constraint
        if not deriv:
            for cons in self.constraints:
                pen += cons(pars, deriv=False)**2
            return self.mu * float(pen)

        # otherwise, compute and return the gradient and hessian
        # along with the quadratic penality
        grad = coo_matrix(([], ([], [])), shape=(1,npars))
        hess = coo_matrix(([], ([],[])), shape=(npars,npars))
        for cons in self.constraints:
            # p=None, because self.model.pars was just updated
            c, dc, d2c = cons(pars, deriv=True)
            pen  += c**2
            # we have restored the true grad convention (-2 -> +2)
            grad += +2. * float(c) * dc
            hess += +2. * dc.T.dot(dc)
            if d2c is not None:
                hess += 2. * c * d2c

        # fixed parameters ?
        idx = pars.indexof() >= 0
        pen = float(pen)
        grad = np.array(grad[:,idx].todense()).squeeze()
        hess = hess[:,idx][idx,:]

        return self.mu * pen, self.mu * grad, self.mu * hess

    def get_rhs(self):
        return np.array([c.rhs for c in self.constraints])



class int_M0_at_phase_cons(LinearConstraint):
    """constraint on the integral of the M0 surface at peak

    This function builds a linear constraint on the M0 parameters.

    .. note:: at this stage, the constraints are a function of the *full*
       parameter vector, not just the free parameters.
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_phase = self.model.basis.by.eval(np.array([self.phase])).toarray()
        pp = pars.copy()
        pp.release()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_phase, gram_dot_filter).ravel())
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        i = np.full(len(C.col), 0)
        j = pp['M0'].indexof(C.col)
        M = coo_matrix((C.data, (i, j)), shape=(1, npars))
        M.data *= (self.model.norm / self.model.int_ab_spec)
        return M


class int_dM0_at_phase_cons(LinearConstraint):
    """constraint on the integral of the phase derivatives of M0 at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_dphase = self.model.basis.by.deriv(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_dphase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M0'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data # / self.model.int_M0_phase_0 # (1.E5 * self.model.int_ab_spec)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= (self.model.norm)
        return M


class int_M1_at_phase_cons(LinearConstraint):
    """constraint on the integral of the M1 surface at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_phase = self.model.basis.by.eval(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_phase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M1'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data   #/ model.int_M0_phase_0
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= self.model.norm
        return M


class int_dM1_at_phase_cons(LinearConstraint):
    """constraint on the integral of the phase derivatives of M1 at peak
    """
    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase
    def init_h_matrix(self, pars):
        J_dphase = self.model.basis.by.deriv(np.array([self.phase])).toarray()
        gram_dot_filter = self.model.get_gram_dot_filter()
        C = coo_matrix(np.outer(J_dphase, gram_dot_filter).ravel())
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['M1'].indexof(C.col)
        i = np.full(len(C.col), 0)
        v = C.data  # / model.int_M0_phase_0
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                        shape=(1, npars))
        M.data *= self.model.norm
        return M


class mean_col_cons(LinearConstraint):
    """constraint on the mean of the color parameters
    """
    def __init__(self, model, rhs):
        super().__init__(model, rhs)
    def init_h_matrix(self, pars):
        nsn = len(pars['X0'].full)
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['c'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars))
        return M


class mean_x1_cons(LinearConstraint):
    """constraint on the mean of the x1 parameters
    """
    def __init__(self, model, rhs):
        super().__init__(model, rhs)
    def init_h_matrix(self, pars):
        nsn = len(pars['X0'].full)
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full)
        j = pp['X1'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        M = coo_matrix((v[idx], (i[idx], j[idx])),
                            shape=(1, npars))
        return M


class x1_var_cons(Constraint):

    def __init__(self, model, rhs):
        self.model = model
        self.rhs = rhs

    def __call__(self, pars, deriv=False):
        """
        """
        # if p is not None:
        #     self.model.pars.free = p.free

        # CHECK: do we need all_pars_released, or just self.model.pars ?
        pp = pars.copy()
        pp.release()
        # pars = self.model.all_pars_released
        npars = len(pp.full) # len(self.model.all_pars_released.full)
        nsn = len(pp['X0'].full)

        # constraint function: h(X1) = \sum_i X1**2
        cons = (pp['X1'].full**2).sum() / nsn  # pp or pars ?
        if not deriv:
            return cons

        # first derivatives of h
        pars.full[:] = 0.
        j = pars['X1'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = pars['X1'].full
        idx = j >= 0
        J = coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars)) # was +1

        # second derivatives of h
        i = pars['X1'].indexof(np.arange(nsn))
        v = np.full(nsn, 2./nsn)
        idx = j >= 0
        H = coo_matrix((v[idx], (i[idx], i[idx])), shape=(npars, npars))

        return cons, J, H


def salt2like_linear_constraints(model, mu=1.E6, Mb=-19.5, dm15=1.): # was 0.96
    """
    """
    m0_0 = int_M0_at_phase_cons(model, 14381300.77605067, phase=0.) # ab_flux_at_10Mpc(Mb=Mb),
    dm0_0 = int_dM0_at_phase_cons(model, 0., phase=0.)
    m1_0 = int_M1_at_phase_cons(model, 0., phase=0.)
    m1_15 = int_M1_at_phase_cons(model, dm15, phase=15.)  # rhs was 0.96
    # dm1 = int_dM1_at_phase_cons(model, 0.)
    col = mean_col_cons(model, 0.)
    x1  = mean_x1_cons(model, 0.)
    # x1_var = x1_var_cons(model, 1.)

    return ConstraintSet([m0_0, dm0_0, m1_0, m1_15, col, x1], mu=mu)


def salt2like_classical_constraints(model, mu=1.E6, Mb=-19.5):
    """
    """
    # m0_0 = int_M0_at_phase_cons(model, ab_flux_at_10Mpc(Mb=Mb), phase=0.)
    m0_0 = int_M0_at_phase_cons(model, 14381300.77605067, phase=0.)
    dm0_0 = int_dM0_at_phase_cons(model, 0., phase=0.)
    m1_0 = int_M1_at_phase_cons(model, 0., phase=0.)
    dm1 = int_dM1_at_phase_cons(model, 0.)
    col = mean_col_cons(model, 0.)
    x1  = mean_x1_cons(model, 0.)
    x1_var = x1_var_cons(model, 1.)

    return ConstraintSet(model,
                         [m0_0, dm0_0, m1_0, dm1,
                          col, x1, x1_var], mu=mu)


# class SALT2LikeConstraints:
#     """This class implements the 7 constraints (6 linear, 1 non-linear)
#     that are used when training a SALT2Like model.


#     It is typically called within the minimize.LogLikelihood class. It
#     returns the value of the contraints (squared, ready to add the to chi2),
#     the contraints gradient and hessian.
#     """

#     # this dict maps the constraint name to the corresponding
#     # block parameter name
#     _blocks = {'M0': 'M0', 'M1': 'M1', 'dM0': 'M0', 'dM1': 'M1',
#                'c': 'c', 'X1': 'X1', 'X1_var': 'X1'}

#     def __init__(self, model, mu=1.E10,
#                  active={'M0': 10**(-0.4 * (30-19.5)),
#                          'M1': 0., 'dM0': 0., 'dM1': 0.,
#                          'c': 0., 'X1': 0., 'X1_var': 1.}):
#         """
#         """
#         self.model = model
#         self.mu = mu
#         self.active = active
#         self.n_linear_cons = self.get_n_linear_cons()

#         # note: self.ncons and self.rhs cannot be known
#         # in advance because some paramters blocks may be fixed
#         # in which case the constraint is disabled
#         # self.ncons = 0
#         self.rhs = []

#         # the constraints act on the full parameter vector.
#         # So, we need to use a copy of it, with all parameters
#         # free to build the constraint matrices.
#         self.pars_full = self.model.pars.copy()
#         self.pars_full.release()
#         # number of parameters (free and fixed) and
#         # number of supernovae
#         self.npars = len(self.pars_full.full)
#         self.nsn = len(self.pars_full['X0'].full)
#         # whether the x1_vars need to be evaluated.
#         self.eval_x1_var = 'X1_var' in self.check_active()

#         # fetch the normalization filter (its projection on the spline basis)
#         # and multiply it with the model gram. This will be needed
#         # to compute the M0, M1, dM0 and dM1 constraints
#         key = self.model.normalization_band_name
#         coeffs, _ = self.model.filter_db[key]
#         self._gram_dot_filter = self.model.gram.dot(coeffs)

#         # and finally, initialize the requested (linear) constraints
#         self._init_constraints()

#     def get_n_cons(self):
#         """
#         """
#         ncons = 0
#         for cons_name in self.check_active():
#             block_name = self._blocks[cons_name]
#             idx = self.model.pars[block_name].indexof() >= 0
#             if idx.sum() > 0:
#                 ncons += 1
#         return ncons

#     def check_active(self):
#         ret = {}
#         for cons_name in self.active:
#             block_name = self._blocks[cons_name]
#             idx = self.model.pars[block_name].indexof() >= 0
#             if idx.sum() > 0:
#                 ret[cons_name] = self.active[cons_name]
#         return ret

#     def get_n_linear_cons(self):
#         ncons = self.get_n_cons()
#         active = self.check_active()
#         if 'X1_var' in active:
#             return len(active) - 1
#         return len(active)

#     def _init_constraints(self):
#         """initialize the 6 linear contraints
#         and the rhs of X1_var (if active)
#         """
#         self.ncons = self.get_n_cons()
#         active = self.check_active()
#         self.rhs = np.zeros(self.ncons)
#         self.Jac = dok_matrix((self.ncons, self.npars))

#         cons_index = 0
#         for cons_name in ['M0', 'M1', 'dM0', 'dM1', 'c', 'X1']:
#             if cons_name not in active:
#                 continue
#             f = getattr(self, 'get_' + cons_name + '_cons_jac')
#             try:
#                 Jac = f(cons_index)
#             except AllParametersFixedError:
#                 logging.warning(f'Warning: {cons_name} disabled ' + \
#                                     'all corresponding parameters are fixed')
#                 continue
#             self.Jac += Jac
#             self.rhs[cons_index] = self.active[cons_name]
#             cons_index += 1
#         if self.eval_x1_var:
#             self.rhs[self.ncons-1] = self.active['X1_var']


#     def get_M0_cons_jac(self, cons_index):
#         """matrix of the constraint on the normalization of M0
#         """
#         # some explanation may be needed:
#         # we first evaluate the bspline basis in the phase-direction
#         J_phase = self.model.basis.by.eval(np.array([0.])).toarray()

#         # we then compute all the cross-products of the phase
#         # functions and the coefficients of the gram x filter_coefficients
#         # and we re-order them so that they match the 2D-basis order
#         # (ravel does the job)
#         # note that some of the non-zero values at the end of this
#         # operation are exceedingly small and should be set to zero.
#         # This comes from the filter projection -- and I think the problem
#         # should be solved at this stage.
#         C = coo_matrix(np.outer(J_phase, self._gram_dot_filter).ravel())

#         # finally, we embed this all in a (ncons x npars) matrix
#         # since all parameters have been freed here, there is no
#         # need to filter out the fixed parameters
#         pars = self.pars_full
#         i = np.full(len(C.col), cons_index)
#         j = pars['M0'].indexof(C.col)
#         M = coo_matrix((C.data, (i, j)),
#                         shape=(self.ncons, self.npars))
#         M.data *= (self.model.norm / self.model.int_ab_spec)
#         return M

#     def get_M1_cons_jac(self, cons_index):
#         """
#         """
#         J_phase = self.model.basis.by.eval(np.array([0.])).toarray()
#         C = coo_matrix(np.outer(J_phase, self._gram_dot_filter).ravel())
#         pars = self.pars_full
#         j = pars['M1'].indexof(C.col)
#         i = np.full(len(C.col), cons_index)
#         v = C.data / self.model.int_M0_phase_0
#         idx = j >= 0
#         M = coo_matrix((v[idx], (i[idx], j[idx])),
#                        shape=(self.ncons, self.npars))
#         M.data *= self.model.norm
#         return M

#     def get_dM0_cons_jac(self, cons_index, dphase=0.): # was -0.25
#         """
#         """
#         # A mon avis, ca deconne dur ici...
#         J_dphase = self.model.basis.by.deriv(np.array([dphase])).toarray()
#         C = coo_matrix(np.outer(J_dphase, self._gram_dot_filter).ravel())

#         pars = self.pars_full
#         j = pars['M0'].indexof(C.col)
#         i = np.full(len(C.col), cons_index)
#         v = C.data # / self.model.int_M0_phase_0 # (1.E5 * self.model.int_ab_spec)
#         idx = j >= 0
#         M = coo_matrix((v[idx], (i[idx], j[idx])),
#                        shape=(self.ncons, self.npars))
#         M.data *= (self.model.norm)
#         return M

#     def get_dM1_cons_jac(self, cons_index):
#         """
#         """
#         J_dphase = self.model.basis.by.deriv(np.array([0.])).toarray()
#         C = coo_matrix(np.outer(J_dphase, self._gram_dot_filter).ravel())
#         pars = self.pars_full
#         j = pars['M1'].indexof(C.col)
#         i = np.full(len(C.col), cons_index)
#         v = C.data / self.model.int_M0_phase_0
#         idx = j >= 0
#         M = coo_matrix((v[idx], (i[idx], j[idx])),
#                        shape=(self.ncons, self.npars))
#         M.data *= self.model.norm
#         return M

#     def get_col_cons_jac(self, cons_index):
#         """
#         """
#         nsn = len(self.model.pars['X0'].full)
#         pars = self.pars_full
#         j = pars['c'].indexof(np.arange(nsn))
#         i = np.full(nsn, cons_index)
#         v = np.full(nsn, 1./nsn)
#         idx = j >= 0
#         return coo_matrix((v[idx], (i[idx], j[idx])),
#                           shape=(self.ncons, self.npars))

#     def get_X1_cons_jac(self, cons_index):
#         nsn = len(self.model.pars['X0'].full)
#         pars = self.pars_full
#         j = pars['X1'].indexof(np.arange(nsn))
#         i = np.full(nsn, cons_index)
#         v = np.full(nsn, 1./nsn)
#         idx = j >= 0
#         return coo_matrix((v[idx], (i[idx], j[idx])),
#                           shape=(self.ncons, self.npars))

#     def eval_x1_var_cons(self, deriv=False):
#         """
#         .. note : we assume that <X1> = 0
#         """
#         assert('X1_var' in self.active)
#         assert len(self.model.pars['X1'].free) > 0

#         pars = self.pars_full
#         nsn = len(pars['X0'].full)

#         # CHECK: are we sure we shouldn't use full instead of free ?
#         cons = (self.model.pars['X1'].full**2).sum() / nsn
#         if not deriv:
#             return cons


#         # first derivatives
#         pars.full[:] = 0.
#         pars['X1'].free = self.model.pars['X1'].full * 2. / nsn

#         j = pars['X1'].indexof(np.arange(nsn))
#         cons_index = self.ncons - 1
#         i = np.full(nsn, cons_index)
#         v = pars['X1'].full
#         idx = j >= 0
#         J = coo_matrix((v[idx], (i[idx], j[idx])),
#                        shape=(self.ncons, self.npars))

#         # second derivatives
#         i = pars['X1'].indexof(np.arange(nsn))
#         v = np.full(nsn, 2./nsn)
#         idx = j >= 0
#         H = coo_matrix((v[idx], (i[idx],i[idx])),
#                        shape=(self.npars, self.npars))

#         return cons, J, H

#     def __call__(self, p, deriv=False):
#         """evaluate the activated constraints and their derivatives
#         """
#         # this is done once for all, with everay __call__
#         # the subfunctions get_XXX or eval_x1_var_cons are not
#         # supposed to update the parameter vector
#         self.model.pars.free = p
#         self.pars_full.full[:] = self.model.pars.full[:]

#         # linear constraints
#         cons = self.Jac @ self.model.pars.full

#         if not deriv:
#             # penalities from the non-linear constraint
#             if self.eval_x1_var:
#                 v  = self.eval_x1_var_cons(deriv=False)
#                 cons[-1] = v
#             cons -= self.rhs
#             return self.mu * (cons**2).sum()

#         H_x1 = None
#         if self.eval_x1_var:
#             v, J, H_x1 = self.eval_x1_var_cons(deriv=True)
#             cons[-1] = v
#             J = J + self.Jac
#         else:
#             J = self.Jac
#         J = J.tocoo()

#         # cons_x1_var = cons[-1]
#         cons -= self.rhs

#         penality = self.mu * (cons**2).sum()
#         jj = self.model.pars.indexof(J.col)
#         idx = jj >= 0
#         n_free_pars = len(self.model.pars.free)
#         JJ = coo_matrix((J.data[idx], (J.row[idx], jj[idx])),
#                         shape=(self.ncons, n_free_pars))
#         grad = -self.mu * 2. * JJ.T.dot(cons)
#         hess =  self.mu * 2. * JJ.T.dot(JJ)
#         # if H_x1 is not None:
#         #     H_x1 = H_x1.tocsr()
#         #     idx = self.model.pars.indexof() >= 0
#         #     hess += (self.mu * 2. * cons[-1] * H_x1[idx,:][:,idx])

#         return penality, grad, hess

#     def get_linearized_constraints(self, p):
#         """return the linearized constraints

#         .. note:: full parameter size !
#         """
#         _, J, _ = self.eval_x1_var_cons(deriv=True)
#         J = J + self.Jac
#         return J

#     def get_rhs(self):
#         """
#         """
#         return self.rhs

# # original class written by Guy
# # it has been rewritten above, to simplify the handling of parameters
# class Constraints2D:
#     r"""
#     Local (sparse) constraints to apply to the spline part of the salt model.
#     The SED model is defined as :

#      .. math::
#         S(\lambda, t) = X_0 [M_0(\lambda, p) + X_1 M_1(\lambda, p)] \, e^{CL(\lambda) \, C}

#     where the 2D surfaces are defined on the same spline basis :

#     .. math::
#         M_{0|1}(\lambda, \mathrm{p}) = \sum_{k\ell} \theta_{k\ell}^{0|1} \mathcal{B}^t_k(\mathrm{p})
#         \mathcal{B}^m_l(\mathrm{\lambda})

#     Constraints are:

#     .. math::
#         \int M_0(\lambda, p = 0, ) T_B(\lambda) \frac{\lambda}{hc} d\lambda = 1 \\

#          \int M_1(\lambda, p = 0, ) T_B(\lambda) \frac{\lambda}{hc} d\lambda = 1 \\

#         \int \left.\frac{\partial M_0(\lambda, p = 0)}{\partial t}\right|_{t = t_{max}^{B^\star}}
#         T_B(\lambda) \frac{\lambda}{hc} d\lambda = 0 \\

#          \int \left.\frac{\partial M_1(\lambda, p = 0)}{\partial t}\right|_{t = t_{max}^{B^\star}}
#          T_B(\lambda) \frac{\lambda}{hc} d\lambda = 0 \\

#          \left<c\right> = 0 \\

#          \left<X_1\right> = 0 \\

#          \left<(X_1 - \left<X_1\right>)^2\right> = 1 \\

#     Six first constraints are linear. They are implemented in the same form :

#     .. math::
#         C(\beta) =  \alpha - H_{pen} \beta

#     For practical reasons, :math:`H_{pen}` and :math:`\alpha` are divided by the square root of the constraints
#     quadratic penalty, :math:`\sigma_{pen} = \sqrt{\mu_{pen}}`

#     The last one is a quadratic constraint then it is called : :math:`C^1(X_1)`
#     Only the constraint involving the free parameters of the model are considered.

#     Attributes
#     ----------
#     model : nacl.salt
#         Model.
#     pars : nacl.lib.fitparameters
#         Model parameters.
#     basis : nacl.lib.bspline
#         Model basis 2D surfaces splines 2D basis (wavelength, phase)

#     parameters_cons : list
#         Free parameters of the model, parameters to constraints

#     x1_var_real : bool
#         If False, consider that the mean of the constraint on the :math:`X_1` is already true, which mean that

#          .. math::
#              C^1(X_1) = \left<(X_1)^2\right> = 1

#         If True :

#          .. math::
#              C^1(X_1) = \left<(X_1 - \left<X_1\right>)^2\right> = 1

#     vals : array
#          Values of the :math:`\alpha` vector, divided by :math:`\sigma_{pen}`

#     sig_vals : array
#         Values of the :math:`\sigma_{pen}` vector.

#     mu_pq : float
#         Values of the linear constraint quadratic penalty amplitude.
#     mu_pq_var_x1 :
#         Values of the :math:`X_1` variance constraint quadratic penalty amplitude.

#     H : scipy.sparse.coo_matrix
#         :math:`H_{pen}` divided by the square root of the constraints quadratic penalty, :math:`\sigma_{pen}`
#     """

#     def __init__(self, model, mu_pq, parameters_cons=['M0', 'X0', 'tmax'],
#                  mu_pq_var_x1=None, x1_var_real=False):
#         r"""
#         Constructor.
#         Creation of the sparse matrix :math:`H_{pen}`.
#         Since the constraints are linear, each line correspond to a constraint and the colon to the model parameters
#         (: :math:`\beta`) involved in this constraint.

#         Parameters
#         ----------
#         model : nacl.salt
#             Model.
#         mu_pq : float
#             Values of the linear constraint quadratic penalty amplitude.
#         parameters_cons : list
#             Free parameters of the model, parameters to constraints
#         mu_pq_var_x1 :
#             Values of the :math:`X_1` variance constraint quadratic penalty amplitude.
#         x1_var_real : bool
#             If False, consider that the mean of the constraint on the :math:`X_1` is already true, which mean that

#              .. math::
#                  C^1(X_1) = \left< (X_1)^2 \right> = 1

#             If True :

#              .. math::
#                  C^1(X_1) = \left< (X_1 - \left< X_1 \right> )^2 \right> = 1
#         """
#         self.model = model
#         self.parameters_cons = parameters_cons
#         self.x1_var_real = x1_var_real

#         self.pars = self.model.pars
#         self.basis = self.model.basis

#         n = len(self.pars.full)
#         pp = self.pars.copy()
#         pp.release()
#         nsn = len(pp['X0'].full)
#         self.vals = []  # vals
#         tr = self.model.filter_db.transmission_db['SWOPE::B']
#         tr_eval, _ = self.model.filter_db.insert(tr, z=0)
#         filter_z = self.model.gram.dot(tr_eval)
#         comp_t = 0
#         self.mu_pq = mu_pq
#         if mu_pq_var_x1 is None:
#             self.mu_pq_var_x1 = self.mu_pq
#         else:
#             self.mu_pq_var_x1 = mu_pq_var_x1
#         sigma = 1/np.sqrt(self.mu_pq)
#         self.sig_vals = []

#         i = np.array([])
#         j = np.array([])
#         v = np.array([])

#         if ('M0' in parameters_cons) & ('X0' in parameters_cons):
#             # first constraint: S(phase = 0, band = B) == 1.
#             surface0_eval = self.model.basis.by.eval(np.array([0.])).toarray()
#             jacobian = coo_matrix(np.outer(surface0_eval, filter_z).ravel())
#             i = np.hstack((i, pp['M0'].indexof(jacobian.col)))
#             j = np.hstack((j, np.full(len(jacobian.col), comp_t)))
#             v = np.hstack((v, jacobian.data.copy()))
#             comp_t += 1
#             self.vals.append(1)
#             self.sig_vals.append(sigma)

#         if ('M0' in parameters_cons) & ('tmax' in parameters_cons):
#             # second constraint: S'(phase = 0) == 0. band = B
#             surface_zero_derivative = self.model.basis.by.deriv(np.array([0.])).toarray()
#             integral_zero_derivative = coo_matrix(np.outer(surface_zero_derivative, filter_z).ravel())
#             i = np.hstack((i, pp['M0'].indexof(integral_zero_derivative.col)))
#             j = np.hstack((j, np.full(len(integral_zero_derivative.col), comp_t)))
#             v = np.hstack((v, integral_zero_derivative.data.copy()))
#             comp_t += 1
#             self.vals.append(0)
#             self.sig_vals.append(sigma)

#         if ('M1' in parameters_cons) & ('X1' in parameters_cons):
#             # third constraint: flux_broadband_max  = 0
#             surface1_eval = self.model.basis.by.eval(np.array([0.])).toarray()
#             jacobian = coo_matrix(np.outer(surface1_eval, filter_z).ravel())
#             i = np.hstack((i, pp['M1'].indexof(jacobian.col)))
#             j = np.hstack((j, np.full(len(jacobian.col), comp_t)))
#             v = np.hstack((v, jacobian.data.copy()))
#             comp_t += 1
#             self.vals.append(0)
#             self.sig_vals.append(sigma)

#         if ('M1' in parameters_cons) & ('tmax' in parameters_cons):
#             # fourth constraint: flux_broadband_max = 0
#             surface1_zero_derivative = self.model.basis.by.deriv(np.array([0.])).toarray()
#             integral1_zero_derivative = coo_matrix(np.outer(surface1_zero_derivative, filter_z).ravel())
#             i = np.hstack((i, pp['M1'].indexof(integral1_zero_derivative.col)))
#             j = np.hstack((j, np.full(len(integral1_zero_derivative.col), comp_t)))
#             v = np.hstack((v, integral1_zero_derivative.data.copy()))
#             comp_t += 1
#             self.vals.append(0)
#             self.sig_vals.append(sigma)

#         v *= self.model.norm

#         if 'X1' in parameters_cons:
#             # mean X1 = 0
#             i = np.hstack((i, pp['X1'].indexof(np.arange(nsn))))
#             j = np.hstack((j, np.full(nsn, comp_t)))
#             v = np.hstack((v, np.full(nsn, 1)))  # X1 new constraint 1e-5
#             comp_t += 1
#             self.vals.append(0)
#             self.sig_vals.append(sigma)

#         if 'c' in parameters_cons:
#             # mean c = 0
#             i = np.hstack((i, pp['c'].indexof(np.arange(nsn))))
#             j = np.hstack((j, np.full(nsn, comp_t)))
#             v = np.hstack((v, np.full(nsn, 1)))
#             comp_t += 1
#             self.vals.append(0)
#             self.sig_vals.append(sigma)

#         self.cl_i = None
#         self.sig_vals = np.array(self.sig_vals)
#         self.vals = np.array(self.vals)/self.sig_vals
#         v /= self.sig_vals[np.array(j).astype(int)]

#         self.H = coo_matrix((v, (i, j)), shape=(n, len(self.vals))).tocsr()

#         if 'X1' in self.parameters_cons:
#             self.var_x1 = self.var_x1
#         else:
#             self.var_x1 = None

#     def var_x1(self, beta, gamma=[], jac=False):
#         r"""
#         Evaluate the seventh constraints, on the variance of the :math:`X_1` parameters.

#         If the attributes x1_var_real is True, then compute :
#             the constraint :

#             .. math::
#                  C^1(X_1) = \left<(X_1 - \left<X_1\right>)^2\right> = 1

#             it firsts derivatives wrt :math:`X_1` :

#             .. math::
#                 \frac{dC(X_1)}{dX_1^i} = \frac{2}{N} (X_1 - \bar{X_1})^T \left(\begin{matrix}
#                 - \frac{1}{N} \\
#                 ... \\
#                 (1 - \frac{1}{N})_i \\
#                 ... \\
#                 - \frac{1}{N}
#                 \end{matrix} \right)

#             it seconds derivatives wrt :math:`X_1` :

#                 .. math::
#                     \frac{d^2C(X_1)}{dX_1^j dX_1^i} &= \frac{2}{N} \left(\begin{matrix}
#                     - \frac{1}{N} \\
#                     ... \\
#                     (1 - \frac{1}{N})_j \\
#                     ... \\
#                     - \frac{1}{N}
#                     \end{matrix} \right)^T  \left(\begin{matrix}
#                     - \frac{1}{N} \\
#                     ... \\
#                     (1 - \frac{1}{N})_i \\
#                     ... \\
#                     - \frac{1}{N}
#                     \end{matrix} \right)

#         If the attributes x1_var_real is False :

#         .. math::
#             C^1(X_1) =  \frac{1}{N} X_1^T X_1 \quad  \& \quad
#             \frac{\partial C^1(X_1)}{\partial X_1^i} = \frac{2}{N^2} X_1^T  \quad  \& \quad
#             \frac{\partial^2 C^1(X_1)}{ \partial X_1^j \partial X_1^i} = \frac{2}{N^2} \delta{ij}


#         Parameters
#         ----------
#         beta : array
#             Model parameters.
#         gamma : array
#             Variance model parameters.
#         jac : bool
#             If derivatives are needed.

#         Returns
#         -------
#         cons : array
#             Value of the constraints multiply by the square root of mu_pq_var_x1

#         if jac is True :
#         jacobian_cons : array
#             Gradient of the constraints multiply by mu_pq_var_x1
#         hessian_cons :
#             Hessian of the constraints multiply by mu_pq_var_x1
#         """

#         self.pars.free = beta
#         pp = self.pars.copy()
#         nsn = len(pp['X1'].full)
#         diag = np.zeros_like(pp.free)
#         i = pp['X1'].indexof(np.arange(nsn))
#         xh, yh = np.meshgrid(i, i)
#         if self.x1_var_real is False:
#             diag[i] = np.ones(nsn)
#             cons = pp['X1'].free.var() - 1
#             dc0 = 2/nsn * diag * pp.free
#             dc = np.matrix(dc0[i])
#             jacobian_cons = 2 * dc0 * cons
#             # if gamma is not None:
#             # jacobian_cons = self.mu_pq * np.hstack((jacobian_cons, np.zeros(len(gamma))))
#             jacobian_cons = self.mu_pq_var_x1 * np.hstack((jacobian_cons, np.zeros(len(gamma))))
#             hh = np.diag([2/nsn] * len(i))
#             # hessian_cons = self.mu_pq * (2 * np.matrix(hh) * cons + 2 * dc.T @ dc)
#             hessian_cons = self.mu_pq_var_x1 * (2 * np.matrix(hh) * cons + 2 * dc.T @ dc)

#         # ## new implementation
#         # ## here mean is not Zero, mush longer for the fit to
#         # ## reach the constraints but when we start far from
#         # ## <X1> != 0 needed
#         elif self.x1_var_real:
#             jacobian_cons = np.zeros_like(pp.free)
#             x1mean = pp['X1'].free.mean()
#             x1v = pp['X1'].free - x1mean
#             cons = pp['X1'].free.var() - 1  # np.mean(x1v**2) - 1

#             dc_dxi = 2 / nsn ** 2 * (nsn * pp['X1'].free - (2 * nsn - 1) * x1mean)
#             jacobian_cons[i] = 2 * cons * dc_dxi
#             jacobian_cons = self.mu_pq_var_x1 * \
#                 np.hstack((jacobian_cons, np.zeros(len(gamma))))

#             dc_dxi_2 = 2 / nsn ** 2 * (nsn * np.diag(np.ones_like(x1v)) - \
#                                        (2 * nsn - 1) / nsn)
#             hessian_cons = self.mu_pq_var_x1 * 2 * (dc_dxi.reshape(-1, 1) @ \
#                                                     dc_dxi.reshape(-1, 1).T + \
#                                                     cons * dc_dxi_2)

#         idx = (yh.ravel() != -1) & (xh.ravel() != -1)
#         hessian_cons = coo_matrix((np.array(hessian_cons).ravel()[idx],
#                                    (yh.ravel()[idx], xh.ravel()[idx])),
#                                   shape=(len(pp.free)+len(gamma), len(pp.free)+len(gamma)))
#         cons *= np.sqrt(self.mu_pq_var_x1)
#         if jac:
#             return cons, jacobian_cons, hessian_cons
#         return cons

#     def get_rhs(self):
#         i = np.where(self.pars.indexof() >= 0)[0]
#         h = self.H[i, :].tocoo()
#         idx = np.bincount(h.col, minlength=len(self.vals)) > 0
#         return self.vals[idx]

#     def __call__(self, beta, jac=False):
#         r"""
#         Evaluate the linear constraints: :math:`C(\beta) = \alpha - H_{pen} \beta`

#         Parameters
#         ----------
#         beta : array
#             Model parameters.
#         jac : bool
#             If derivatives are needed.

#         Returns
#         -------
#         v : array
#             Value of the constraints multiply by the square root of :math:`\mu_{pq}`
#         if jac is True :
#         h : array
#             Matrix :math:`H_{pen}` by :math:`\mu_{pq}`
#         """
#         self.pars.free = beta
#         v = -self.H.T @ self.pars.full + self.vals
#         # we need to evaluate how many constraints are still valid,
#         # given which parameters are fixed. We do that by slicing the
#         # constraint matrix. There is probably a better and faster
#         # way, which avoids having to do this each time the
#         # constraints are evaluated
#         i = np.where(self.pars.indexof() >= 0)[0]
#         h = self.H[i, :].tocoo()
#         # verify that all columns of H are populated
#         j = np.where(np.bincount(h.col) > 0)[0]
#         h = h.tocsr()[:, j].tocoo()
#         # we need also to slice the return values
#         idx = np.bincount(h.col, minlength=len(self.vals)) > 0
#         if jac == 1:
#             return v[idx], h.T
#         return v[idx]



#!/usr/bin/env python3

import numpy as np
import pylab as pl

from nacl.models.toy1d import lightcurves as lc1d
from nacl.models.toy2d import lightcurves as lc2d
from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer

from nacl.models.helpers import check_grad, check_deriv_old
from nacl.lib.fitparameters import FitParameters
from scipy.optimize import curve_fit

from sksparse import cholmod


def constant(x, c):
    return c

def test_new_noise(n=10,nsn=200, npts=20, nbands=3,
                       release=['x0', 'stretch', 'color', 'tmax', 'eta'],
                       gen_error_pedestal=0., gen_error_snake=0.1,
                       gamma_init=0.):
    gen = lc1d.ToyModelGenerator(nbands=nbands, nsn=nsn, npts=npts, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
        
    dp = gen.generate()
    idx = dp.sn_index == 0
    
    pl.figure(figsize=(10,7))
    
    for i in range(n):
        
        pl.errorbar(dp.x[idx], dp.y[idx], yerr = dp.yerr[idx], ls='', marker='.', elinewidth = 2) 
        
        gen.change_noise(dp)
    pl.scatter(dp.x[idx], dp.y_true[idx], color = 'r', marker='x', zorder=n+1, label='y_true')
    pl.xlabel('x')
    pl.ylabel('y')
    pl.legend()
    pl.title('1st LC simulated ' + str(n) + ' times with different noise')
    
    
def bias_1d(n=50,nsn=100, npts=20, nbands=3,
                           release=['x0', 'stretch', 'color', 'tmax', 'eta'],
                           gen_error_pedestal=0., gen_error_snake=0.1,
                           gamma_init=0.05):
    
    x0 = []
    stretch = []
    color = []
    
    gen = lc1d.ToyModelGenerator(nbands=nbands, nsn=nsn, npts=npts, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
        
    dp = gen.generate()
    
    for i in range(n):        
        model = lc1d.LightCurveModel1DWithCalibScatter(dp)
        reg = lc1d.Regularization(block_name='theta')
        cons = lc1d.cons(model, mu=1.E6, s=True)
        snake = lc1d.SimpleErrorSnake(model)
        V_calib = np.diag(np.full(nbands, 0.005**2))
        calib_prior = lc1d.CalibPrior(V_calib)
        ll = LogLikelihood(model,
                           reg=[reg], cons=[cons],
                           priors=[calib_prior],
                           data=dp)
        # fit initialization
        for block_name in ['x0', 'stretch', 'color', 'tmax']:
            ll.pars[block_name].full[:] = dp.sample[block_name]
            ll.pars[block_name].fix()

        for block_name in release:
            ll.pars[block_name].release()

        # first fit with model only
        minz = Minimizer(ll)
        r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)

        # second fit, with model and error modelfull )
        ll = LogLikelihood(model, variance_model=snake,
                           reg=[reg], cons=[cons],
                           priors=[calib_prior],
                           data=dp)
        for block_name in ['x0', 'tmax', 'stretch', 'color', 'theta']:
            ll.pars[block_name].full[:] = r['pars'][block_name].full[:]
        ll.pars['gamma'].full[:] = gamma_init
        minz = Minimizer(ll)
        r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)

        minz.gen = gen
        
        x0.append( minz.log_likelihood.pars['x0'].full )
        stretch.append( minz.log_likelihood.pars['stretch'].full )
        color.append( minz.log_likelihood.pars['color'].full )
        
        gen.change_noise(dp)
    # PLOTS and VARIANCES
    x0 = np.vstack(x0)
    stretch = np.vstack(stretch)
    color = np.vstack(color)
    
    mean_x0 = np.mean(x0, axis=0)
    mean_stretch = np.mean(stretch, axis=0)
    mean_color = np.mean(color, axis=0)
    
    var_x0 = np.var(x0, axis=0)/n
    var_stretch = np.var(stretch, axis=0)/n
    var_color = np.var(color, axis=0)/n
    
    x = np.arange(len(mean_x0))
    
    #empirical covariance
    V_x0_emp = (x0 - mean_x0).T @ (x0 - mean_x0) / (n-1)
    V_stretch_emp = (stretch - mean_stretch).T @ (stretch - mean_stretch) / (n-1)
    V_color_emp = (color - mean_color).T @ (color - mean_color) / (n-1)
    
    #fisher covariance
    free_pars = ['x0', 'stretch', 'color']
    V, err_of_interest, corr = minz.get_cov_matrix(free_pars, corr=True, plot=False)
    idx_x0 = minz.log_likelihood.pars.indexof('x0')
    idx_stretch = minz.log_likelihood.pars.indexof('stretch')
    idx_color = minz.log_likelihood.pars.indexof('color')
    
    V_x0_fisher = V.toarray()[idx_x0][:,idx_x0]
    V_stretch_fisher = V.toarray()[idx_stretch][:,idx_stretch]
    V_color_fisher = V.toarray()[idx_color][:,idx_color]
    
    #fit the bias
    popt_x0, pcov_x0 = curve_fit(constant, x,  dp.sample['x0'] - mean_x0)
    popt_stretch, pcov_stretch = curve_fit(constant, x,  dp.sample['stretch'] - mean_stretch)
    popt_color, pcov_color = curve_fit(constant, x,  dp.sample['color'] - mean_color)
    
    
    #plot true vs fit differences and pulls
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].errorbar(x, dp.sample['x0'] - mean_x0, np.sqrt(var_x0), marker = '.', color ='k', ls='')
    axes[0].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[0].plot(x, np.linspace(popt_x0[0], popt_x0[0], len(x)), 'g--', label='bias fit')
    axes[0].fill_between(x, np.linspace(popt_x0[0]-np.sqrt(pcov_x0[0][0]), popt_x0[0]-np.sqrt(pcov_x0[0][0]), 
                         len(x)), np.linspace(popt_x0[0]+np.sqrt(pcov_x0[0][0]), popt_x0[0]+np.sqrt(pcov_x0[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[0].set_title('x0 true - x0 fit')
    axes[0].set_ylabel(r'$x0_{true}$ - $x0_{fit}$')
    axes[0].legend()
    
    axes[1].errorbar(x, dp.sample['stretch'] - mean_stretch, np.sqrt(var_stretch), marker = '.', color ='k', ls='')
    axes[1].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[1].plot(x, np.linspace(popt_stretch[0], popt_stretch[0], len(x)), 'g--', label='bias fit')
    axes[1].fill_between(x, np.linspace(popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), 
                         len(x)), np.linspace(popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[1].set_title('stretch true - stretch fit')
    axes[1].set_ylabel(r'$stretch_{true}$ - $stretch_{fit}$')
    axes[1].legend()
    
    axes[2].errorbar(x, dp.sample['color'] - mean_color, np.sqrt(var_color), marker = '.', color ='k', ls='')
    axes[2].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[2].plot(x, np.linspace(popt_color[0], popt_color[0], len(x)), 'g--', label='bias fit')
    axes[2].fill_between(x, np.linspace(popt_color[0]-np.sqrt(pcov_color[0][0]), popt_color[0]-np.sqrt(pcov_color[0][0]), 
                         len(x)), np.linspace(popt_color[0]+np.sqrt(pcov_color[0][0]), popt_color[0]+np.sqrt(pcov_color[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[2].set_title('color true - color fit')
    axes[2].set_ylabel(r'$color_{true}$ vs $color_{fit}$')
    axes[2].legend()
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x, np.sqrt(var_x0) - np.sqrt(np.diag(V_x0_fisher)), 'r.')
    axes[0].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[0].set_title('err_x0 empirical - err_x0 fisher')
    axes[0].set_ylabel(r'err_$x0_{empirical}$ - err_$x0_{fisher}$')
    
    axes[1].plot(x, np.sqrt(var_stretch) - np.sqrt(np.diag(V_stretch_fisher)), 'g.')
    axes[1].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[1].set_title('err_stretch empirical - err_stretch fisher')
    axes[1].set_ylabel(r'err_$stretch_{empirical}$ - err_$stretch_{fisher}$')
    
    axes[2].plot(x, np.sqrt(var_color) - np.sqrt(np.diag(V_color_fisher)), 'b.')
    axes[2].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[2].set_title('err_color empirical - err_color fisher')
    axes[2].set_ylabel(r'err_$color_{empirical}$ - err_$color_{fisher}$')
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x, np.abs(dp.sample['x0'] - mean_x0)/np.sqrt(var_x0), 'k.')
    axes[0].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[0].set_title('x0 pull')
    axes[0].set_ylabel(r'|$x0_{true}$ - $x0_{fit}$|/err_x0')
    
    axes[1].plot(x, np.abs(dp.sample['stretch'] - mean_stretch)/np.sqrt(var_stretch), 'k.')
    axes[1].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[1].set_title('stretch pull')
    axes[1].set_ylabel(r'|$stretch_{true}$ - $stretch_{fit}$|/err_stretch')
    
    axes[2].plot(x, np.abs(dp.sample['color'] - mean_color)/np.sqrt(var_color), 'k.')
    axes[2].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[2].set_title('color pull')
    axes[2].set_ylabel(r'|$color_{true}$ - $color_{fit}$|/err_color')
    
    
    #remove outliers
    idx_x0_1 = dp.sample['x0'] - mean_x0 - np.sqrt(var_x0) < np.linspace(popt_x0[0], popt_x0[0], len(x)) + np.linspace(np.sqrt(pcov_x0[0][0]), np.sqrt(pcov_x0[0][0]), len(x))
    idx_x0_2 = dp.sample['x0'][idx_x0_1] - mean_x0[idx_x0_1] + np.sqrt(var_x0)[idx_x0_1] > np.linspace(popt_x0[0], popt_x0[0], len(x))[idx_x0_1] - np.linspace(np.sqrt(pcov_x0[0][0]), np.sqrt(pcov_x0[0][0]), len(x))[idx_x0_1]
    idx_x0_3 = np.abs(dp.sample['x0'][idx_x0_1][idx_x0_2] - mean_x0[idx_x0_1][idx_x0_2])<1
    
    idx_stretch_1 = dp.sample['stretch'] - mean_stretch - np.sqrt(var_stretch) < np.linspace(popt_stretch[0], popt_stretch[0], len(x)) + np.linspace(np.sqrt(pcov_stretch[0][0]), np.sqrt(pcov_stretch[0][0]), len(x))
    idx_stretch_2 = dp.sample['stretch'][idx_stretch_1] - mean_stretch[idx_stretch_1] + np.sqrt(var_stretch)[idx_stretch_1] > np.linspace(popt_stretch[0], popt_stretch[0], len(x))[idx_stretch_1] - np.linspace(np.sqrt(pcov_stretch[0][0]), np.sqrt(pcov_stretch[0][0]), len(x))[idx_stretch_1]
    idx_stretch_3 = np.abs(dp.sample['stretch'][idx_stretch_1][idx_stretch_2] - mean_stretch[idx_stretch_1][idx_stretch_2])<0.05

    idx_color_1 = dp.sample['color'] - mean_color - np.sqrt(var_color) < np.linspace(popt_color[0], popt_color[0], len(x)) + np.linspace(np.sqrt(pcov_color[0][0]), np.sqrt(pcov_color[0][0]), len(x))
    idx_color_2 = dp.sample['color'][idx_color_1] - mean_color[idx_color_1] + np.sqrt(var_color)[idx_color_1] > np.linspace(popt_color[0], popt_color[0], len(x))[idx_color_1] - np.linspace(np.sqrt(pcov_color[0][0]), np.sqrt(pcov_color[0][0]), len(x))[idx_color_1]
    idx_color_3 = np.abs(dp.sample['color'][idx_color_1][idx_color_2] - mean_color[idx_color_1][idx_color_2])<1
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].errorbar(x[idx_x0_1][idx_x0_2][idx_x0_3], dp.sample['x0'][idx_x0_1][idx_x0_2][idx_x0_3] - mean_x0[idx_x0_1][idx_x0_2][idx_x0_3], np.sqrt(var_x0)[idx_x0_1][idx_x0_2][idx_x0_3], marker = '.', color ='k', ls='')
    axes[0].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[0].plot(x, np.linspace(popt_x0[0], popt_x0[0], len(x)), 'g--', label='bias fit')
    axes[0].fill_between(x, np.linspace(popt_x0[0]-np.sqrt(pcov_x0[0][0]), popt_x0[0]-np.sqrt(pcov_x0[0][0]), 
                         len(x)), np.linspace(popt_x0[0]+np.sqrt(pcov_x0[0][0]), popt_x0[0]+np.sqrt(pcov_x0[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[0].set_title('x0 true - x0 fit no outliers')
    axes[0].set_ylabel(r'$x0_{true}$ - $x0_{fit}$')
    axes[0].legend()
    
    axes[1].errorbar(x[idx_stretch_1][idx_stretch_2][idx_stretch_3], dp.sample['stretch'][idx_stretch_1][idx_stretch_2][idx_stretch_3] - mean_stretch[idx_stretch_1][idx_stretch_2][idx_stretch_3], np.sqrt(var_stretch)[idx_stretch_1][idx_stretch_2][idx_stretch_3], marker = '.', color ='k', ls='')
    axes[1].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[1].plot(x, np.linspace(popt_stretch[0], popt_stretch[0], len(x)), 'g--', label='bias fit')
    axes[1].fill_between(x, np.linspace(popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), 
                         len(x)), np.linspace(popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[1].set_title('stretch true - stretch fit no outliers')
    axes[1].set_ylabel(r'$stretch_{true}$ - $stretch_{fit}$')
    axes[1].legend()
    
    axes[2].errorbar(x[idx_color_1][idx_color_2][idx_color_3], dp.sample['color'][idx_color_1][idx_color_2][idx_color_3] - mean_color[idx_color_1][idx_color_2][idx_color_3], np.sqrt(var_color)[idx_color_1][idx_color_2][idx_color_3], marker = '.', color ='k', ls='')
    axes[2].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[2].plot(x, np.linspace(popt_color[0], popt_color[0], len(x)), 'g--', label='bias fit')
    axes[2].fill_between(x, np.linspace(popt_color[0]-np.sqrt(pcov_color[0][0]), popt_color[0]-np.sqrt(pcov_color[0][0]), 
                        len(x)), np.linspace(popt_color[0]+np.sqrt(pcov_color[0][0]), popt_color[0]+np.sqrt(pcov_color[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[2].set_title('color true - color fit no outliers')
    axes[2].set_ylabel(r'$color_{true}$ - $color_{fit}$')
    axes[2].legend()
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x[idx_x0_1][idx_x0_2][idx_x0_3], np.sqrt(var_x0[idx_x0_1][idx_x0_2][idx_x0_3]) - np.sqrt(np.diag(V_x0_fisher))[idx_x0_1][idx_x0_2][idx_x0_3], 'r.')
    axes[0].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[0].set_title('err_x0 empirical - err_x0 fisher no outliers')
    axes[0].set_ylabel(r'err_$x0_{empirical}$ - err_$x0_{fisher}$')
    
    axes[1].plot(x[idx_stretch_1][idx_stretch_2][idx_stretch_3], np.sqrt(var_stretch[idx_stretch_1][idx_stretch_2][idx_stretch_3]) - np.sqrt(np.diag(V_stretch_fisher)[idx_stretch_1][idx_stretch_2][idx_stretch_3]), 'g.')
    axes[1].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[1].set_title('err_stretch empirical - err_stretch fisher no outliers')
    axes[1].set_ylabel(r'err_$stretch_{empirical}$ - err_$stretch_{fisher}$')
    
    axes[2].plot(x[idx_color_1][idx_color_2][idx_color_3], np.sqrt(var_color[idx_color_1][idx_color_2][idx_color_3]) - np.sqrt(np.diag(V_color_fisher)[idx_color_1][idx_color_2][idx_color_3]), 'b.')
    axes[2].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[2].set_title('err_color empirical - err_color fisher no outliers')
    axes[2].set_ylabel(r'err_$color_{empirical}$ - err_$color_{fisher}$')
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x[idx_x0_1][idx_x0_2][idx_x0_3], np.abs(dp.sample['x0'][idx_x0_1][idx_x0_2][idx_x0_3] - mean_x0[idx_x0_1][idx_x0_2][idx_x0_3])/np.sqrt(var_x0[idx_x0_1][idx_x0_2][idx_x0_3]), 'k.')
    axes[0].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[0].set_title('x0 pull outliers')
    axes[0].set_ylabel(r'|$x0_{true}$ - $x0_{fit}$|/err_x0')
    
    axes[1].plot(x[idx_stretch_1][idx_stretch_2][idx_stretch_3], np.abs(dp.sample['stretch'][idx_stretch_1][idx_stretch_2][idx_stretch_3] - mean_stretch[idx_stretch_1][idx_stretch_2][idx_stretch_3])/np.sqrt(var_stretch[idx_stretch_1][idx_stretch_2][idx_stretch_3]), 'k.')
    axes[1].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[1].set_title('stretch pull no outliers')
    axes[1].set_ylabel(r'|$stretch_{true}$ - $stretch_{fit}$|/err_stretch')
    
    axes[2].plot(x[idx_color_1][idx_color_2][idx_color_3], np.abs(dp.sample['color'][idx_color_1][idx_color_2][idx_color_3] - mean_color[idx_color_1][idx_color_2][idx_color_3])/np.sqrt(var_color[idx_color_1][idx_color_2][idx_color_3]), 'k.')
    axes[2].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[2].set_title('color pull no outliers')
    axes[2].set_ylabel(r'|$color_{true}$ - $color_{fit}$|/err_color')
    #empirical correlation matrix
    v_x0 = np.diag(V_x0_emp)
    i = np.arange(len(v_x0))
    ii = np.vstack([i for k in range( len(v_x0) )])
    v1_x0 = v_x0[ii]
    v2_x0 = v1_x0.T
    vii_jj_x0 = v1_x0 * v2_x0
    vii_jj_x0 = np.sqrt(vii_jj_x0)
    corr_x0_emp = V_x0_emp/vii_jj_x0   
    
    v_stretch = np.diag(V_stretch_emp)
    i = np.arange(len(v_stretch))
    ii = np.vstack([i for k in range( len(v_stretch) )])
    v1_stretch = v_stretch[ii]
    v2_stretch = v1_stretch.T
    vii_jj_stretch = v1_stretch * v2_stretch
    vii_jj_stretch = np.sqrt(vii_jj_stretch)
    corr_stretch_emp = V_stretch_emp/vii_jj_stretch   
    
    v_color = np.diag(V_color_emp)
    i = np.arange(len(v_color))
    ii = np.vstack([i for k in range( len(v_color) )])
    v1_color = v_color[ii]
    v2_color = v1_color.T
    vii_jj_color = v1_color * v2_color
    vii_jj_color = np.sqrt(vii_jj_color)
    corr_color_emp = V_color_emp/vii_jj_color
    
    #fisher correlation matrix
    v_x0 = np.diag(V_x0_fisher)
    i = np.arange(len(v_x0))
    ii = np.vstack([i for k in range( len(v_x0) )])
    v1_x0 = v_x0[ii]
    v2_x0 = v1_x0.T
    vii_jj_x0 = v1_x0 * v2_x0
    vii_jj_x0 = np.sqrt(vii_jj_x0)
    corr_x0_fisher = V_x0_fisher/vii_jj_x0   
    
    v_stretch = np.diag(V_stretch_fisher)
    i = np.arange(len(v_stretch))
    ii = np.vstack([i for k in range( len(v_stretch) )])
    v1_stretch = v_stretch[ii]
    v2_stretch = v1_stretch.T
    vii_jj_stretch = v1_stretch * v2_stretch
    vii_jj_stretch = np.sqrt(vii_jj_stretch)
    corr_stretch_fisher = V_stretch_fisher/vii_jj_stretch   
    
    v_color = np.diag(V_color_fisher)
    i = np.arange(len(v_color))
    ii = np.vstack([i for k in range( len(v_color) )])
    v1_color = v_color[ii]
    v2_color = v1_color.T
    vii_jj_color = v1_color * v2_color
    vii_jj_color = np.sqrt(vii_jj_color)
    corr_color_fisher = V_color_fisher/vii_jj_color
    
    #without outliers
    
    #plot covariance matrix
    fig, axes = pl.subplots(figsize=(10,10), nrows=3, ncols=2)
    
    A = axes[0,0].imshow(V_x0_emp)
    pl.colorbar(A)
    axes[0,0].set_title('Covariance matrix x0 empirical')
    
    A = axes[0,1].imshow(V_x0_fisher)
    pl.colorbar(A)
    axes[0,1].set_title('Covariance matrix x0 fisher')
    
    B = axes[1,0].imshow(V_stretch_emp)
    pl.colorbar(B)
    axes[1,0].set_title('Covariance matrix stretch empirical')
    
    B = axes[1,1].imshow(V_stretch_fisher)
    pl.colorbar(B)
    axes[1,1].set_title('Covariance matrix stretch fisher')
    
    C = axes[2,0].imshow(V_color_emp)
    pl.colorbar(C)
    axes[2,0].set_title('Covariance matrix color empirical')
    
    C = axes[2,1].imshow(V_color_fisher)
    pl.colorbar(C)
    axes[2,1].set_title('Covariance matrix color fisher')
    
    #plot correlation matrix
    fig, axes = pl.subplots(figsize=(10,10), nrows=3, ncols=2)
    
    A = axes[0,0].imshow(corr_x0_emp)
    pl.colorbar(A)
    axes[0,0].set_title('Correlation matrix x0 empirical')
    
    A = axes[0,1].imshow(corr_x0_fisher)
    pl.colorbar(A)
    axes[0,1].set_title('Correlation matrix x0 fisher')
    
    B = axes[1,0].imshow(corr_stretch_emp)
    pl.colorbar(B)
    axes[1,0].set_title('Correlation matrix stretch empirical')
    
    B = axes[1,1].imshow(corr_stretch_fisher)
    pl.colorbar(B)
    axes[1,1].set_title('Correlation matrix stretch fisher')
    
    C = axes[2,0].imshow(corr_color_emp)
    pl.colorbar(C)
    axes[2,0].set_title('Correlation matrix color empirical')
    
    C = axes[2,1].imshow(corr_color_fisher)
    pl.colorbar(C)
    axes[2,1].set_title('Correlation matrix color fisher')
    
    
def bias_2d(n=100, nsn=50, nbands=3, npts1=20, npts2=15, release=['x0', 'tmax', 'stretch', 'eta', 'color'], 
                                 gen_error_pedestal=None, gen_error_snake=0.05, gamma_init=0.05):
    
    x0 = []
    stretch = []
    color = []
    
    gen = lc2d.ToyModelGenerator(nbands=nbands, nsn=nsn, npts1=npts1, npts2=npts2, yerr=0.01,
                                        error_pedestal=gen_error_pedestal,
                                        error_snake=gen_error_snake)
    dp = gen.generate()
    for i in range(n):
        model = lc2d.LightCurve2DCalibScatter(dp)
        reg = lc2d.Regularization(block_name='theta')
      
        cons = lc2d.cons(model, mu=1.E6, color='True')
        snake = lc2d.SimpleErrorSnake(model)
        V_calib = np.diag(np.full(nbands, 0.005**2))
        calib_prior = lc2d.CalibPrior(V_calib)
        ll = LogLikelihood(model, reg=[reg], cons=[cons], priors=[calib_prior], data=dp)

        # fit initialization
        for block_name in ['x0', 'tmax', 'stretch', 'color']:
            ll.pars[block_name].full[:] = dp.sample[block_name]
            ll.pars[block_name].fix()

        for block_name in release:
            ll.pars[block_name].release()

        # first fit with model only
        minz = Minimizer(ll)
        r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6, dchi2=10., max_iter=10)
        # second fit, with model and error model
        ll = LogLikelihood(model, variance_model=snake,
                           reg=[reg], cons=[cons],
                           priors=[calib_prior],
                           data=dp)
        for block_name in ['x0', 'tmax', 'stretch', 'color','theta']:
            ll.pars[block_name].full[:] = r['pars'][block_name].full[:]
        ll.pars['gamma'].full[:] = gamma_init
        minz = Minimizer(ll)
        r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
        minz.gen = gen
        
        x0.append( minz.log_likelihood.pars['x0'].full )
        stretch.append( minz.log_likelihood.pars['stretch'].full )
        color.append( minz.log_likelihood.pars['color'].full )
        
        gen.change_noise(dp)
    
    # PLOTS and VARIANCES
    x0 = np.vstack(x0)
    stretch = np.vstack(stretch)
    color = np.vstack(color)
    
    mean_x0 = np.mean(x0, axis=0)
    mean_stretch = np.mean(stretch, axis=0)
    mean_color = np.mean(color, axis=0)
    
    var_x0 = np.var(x0, axis=0)/n
    var_stretch = np.var(stretch, axis=0)/n
    var_color = np.var(color, axis=0)/n
    
    x = np.arange(len(mean_x0))
    
    #empirical covariance
    V_x0_emp = (x0 - mean_x0).T @ (x0 - mean_x0) / (n-1)
    V_stretch_emp = (stretch - mean_stretch).T @ (stretch - mean_stretch) / (n-1)
    V_color_emp = (color - mean_color).T @ (color - mean_color) / (n-1)
    
    #fisher covariance
    free_pars = ['x0', 'stretch', 'color']
    V, err_of_interest, corr = minz.get_cov_matrix(free_pars, corr=True, plot=False)
    idx_x0 = minz.log_likelihood.pars.indexof('x0')
    idx_stretch = minz.log_likelihood.pars.indexof('stretch')
    idx_color = minz.log_likelihood.pars.indexof('color')
    
    V_x0_fisher = V.toarray()[idx_x0][:,idx_x0]
    V_stretch_fisher = V.toarray()[idx_stretch][:,idx_stretch]
    V_color_fisher = V.toarray()[idx_color][:,idx_color]
    
    #fit the bias
    popt_x0, pcov_x0 = curve_fit(constant, x,  dp.sample['x0'] - mean_x0, sigma = np.sqrt(var_x0))
    popt_stretch, pcov_stretch = curve_fit(constant, x,  dp.sample['stretch'] - mean_stretch, sigma = np.sqrt(var_stretch))
    popt_color, pcov_color = curve_fit(constant, x,  dp.sample['color'] - mean_color, sigma = np.sqrt(var_color))
    
    #plot true vs fit differences and pulls
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].errorbar(x, dp.sample['x0'] - mean_x0, np.sqrt(var_x0), marker = '.', color ='k', ls='')
    axes[0].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[0].plot(x, np.linspace(popt_x0[0], popt_x0[0], len(x)), 'g--', label='bias fit')
    axes[0].fill_between(x, np.linspace(popt_x0[0]-np.sqrt(pcov_x0[0][0]), popt_x0[0]-np.sqrt(pcov_x0[0][0]), 
                         len(x)), np.linspace(popt_x0[0]+np.sqrt(pcov_x0[0][0]), popt_x0[0]+np.sqrt(pcov_x0[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[0].set_title('x0 true - x0 fit')
    axes[0].set_ylabel(r'$x0_{true}$ - $x0_{fit}$')
    axes[0].legend()
    
    axes[1].errorbar(x, dp.sample['stretch'] - mean_stretch, np.sqrt(var_stretch), marker = '.', color ='k', ls='')
    axes[1].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[1].plot(x, np.linspace(popt_stretch[0], popt_stretch[0], len(x)), 'g--', label='bias fit')
    axes[1].fill_between(x, np.linspace(popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), popt_stretch[0]-np.sqrt(pcov_stretch[0][0]), 
                         len(x)), np.linspace(popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), popt_stretch[0]+np.sqrt(pcov_stretch[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[1].set_title('stretch true - stretch fit')
    axes[1].set_ylabel(r'$stretch_{true}$ - $stretch_{fit}$')
    axes[1].legend()
    
    axes[2].errorbar(x, dp.sample['color'] - mean_color, np.sqrt(var_color), marker = '.', color ='k', ls='')
    axes[2].plot(x, np.linspace(0,0,len(x)), 'r')
    axes[2].plot(x, np.linspace(popt_color[0], popt_color[0], len(x)), 'g--', label='bias fit')
    axes[2].fill_between(x, np.linspace(popt_color[0]-np.sqrt(pcov_color[0][0]), popt_color[0]-np.sqrt(pcov_color[0][0]), 
                         len(x)), np.linspace(popt_color[0]+np.sqrt(pcov_color[0][0]), popt_color[0]+np.sqrt(pcov_color[0][0]), len(x)), color = 'g',alpha = 0.5)
    axes[2].set_title('color true - color fit')
    axes[2].set_ylabel(r'$color_{true}$ vs $color_{fit}$')
    axes[2].legend()
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x, np.sqrt(var_x0) - np.sqrt(np.diag(V_x0_fisher)), 'r.')
    axes[0].set_title('err_x0 empirical - err_x0 fisher')
    axes[0].set_ylabel(r'err_$x0_{empirical}$ - err_$x0_{fisher}$')
    
    axes[1].plot(x, np.sqrt(var_stretch) - np.sqrt(np.diag(V_stretch_fisher)), 'g.')
    axes[1].set_title('err_stretch empirical - err_stretch fisher')
    axes[1].set_ylabel(r'err_$stretch_{empirical}$ - err_$stretch_{fisher}$')
    
    axes[2].plot(x, np.sqrt(var_color) - np.sqrt(np.diag(V_color_fisher)), 'b.')
    axes[2].set_title('err_color empirical - err_color fisher')
    axes[2].set_ylabel(r'err_$color_{empirical}$ - err_$color_{fisher}$')
    
    fig, axes = pl.subplots(figsize=(7,10), nrows=3, ncols=1)
    
    axes[0].plot(x, np.abs(dp.sample['x0'] - mean_x0)/np.sqrt(var_x0), 'k.')
    axes[0].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[0].set_title('x0 pull')
    axes[0].set_ylabel(r'|$x0_{true}$ - $x0_{fit}$|/err_x0')
    
    axes[1].plot(x, np.abs(dp.sample['stretch'] - mean_stretch)/np.sqrt(var_stretch), 'k.')
    axes[1].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[1].set_title('stretch pull')
    axes[1].set_ylabel(r'|$stretch_{true}$ - $stretch_{fit}$|/err_stretch')
    
    axes[2].plot(x, np.abs(dp.sample['color'] - mean_color)/np.sqrt(var_color), 'k.')
    axes[2].plot(x, np.linspace(1, 1, len(x)), 'g--')
    axes[2].set_title('color pull')
    axes[2].set_ylabel(r'|$color_{true}$ - $color_{fit}$|/err_color')
    
    
    #empirical correlation matrix
    v_x0 = np.diag(V_x0_emp)
    i = np.arange(len(v_x0))
    ii = np.vstack([i for k in range( len(v_x0) )])
    v1_x0 = v_x0[ii]
    v2_x0 = v1_x0.T
    vii_jj_x0 = v1_x0 * v2_x0
    vii_jj_x0 = np.sqrt(vii_jj_x0)
    corr_x0_emp = V_x0_emp/vii_jj_x0   
    
    v_stretch = np.diag(V_stretch_emp)
    i = np.arange(len(v_stretch))
    ii = np.vstack([i for k in range( len(v_stretch) )])
    v1_stretch = v_stretch[ii]
    v2_stretch = v1_stretch.T
    vii_jj_stretch = v1_stretch * v2_stretch
    vii_jj_stretch = np.sqrt(vii_jj_stretch)
    corr_stretch_emp = V_stretch_emp/vii_jj_stretch   
    
    v_color = np.diag(V_color_emp)
    i = np.arange(len(v_color))
    ii = np.vstack([i for k in range( len(v_color) )])
    v1_color = v_color[ii]
    v2_color = v1_color.T
    vii_jj_color = v1_color * v2_color
    vii_jj_color = np.sqrt(vii_jj_color)
    corr_color_emp = V_color_emp/vii_jj_color   
    
    #fisher correlation matrix
    v_x0 = np.diag(V_x0_fisher)
    i = np.arange(len(v_x0))
    ii = np.vstack([i for k in range( len(v_x0) )])
    v1_x0 = v_x0[ii]
    v2_x0 = v1_x0.T
    vii_jj_x0 = v1_x0 * v2_x0
    vii_jj_x0 = np.sqrt(vii_jj_x0)
    corr_x0_fisher = V_x0_fisher/vii_jj_x0   
    
    v_stretch = np.diag(V_stretch_fisher)
    i = np.arange(len(v_stretch))
    ii = np.vstack([i for k in range( len(v_stretch) )])
    v1_stretch = v_stretch[ii]
    v2_stretch = v1_stretch.T
    vii_jj_stretch = v1_stretch * v2_stretch
    vii_jj_stretch = np.sqrt(vii_jj_stretch)
    corr_stretch_fisher = V_stretch_fisher/vii_jj_stretch   
    
    v_color = np.diag(V_color_fisher)
    i = np.arange(len(v_color))
    ii = np.vstack([i for k in range( len(v_color) )])
    v1_color = v_color[ii]
    v2_color = v1_color.T
    vii_jj_color = v1_color * v2_color
    vii_jj_color = np.sqrt(vii_jj_color)
    corr_color_fisher = V_color_fisher/vii_jj_color
    
    #plot covariance matrix
    fig, axes = pl.subplots(figsize=(10,10), nrows=3, ncols=2)
    
    A = axes[0,0].imshow(V_x0_emp)
    pl.colorbar(A)
    axes[0,0].set_title('Covariance matrix x0 empirical')
    
    A = axes[0,1].imshow(V_x0_fisher)
    pl.colorbar(A)
    axes[0,1].set_title('Covariance matrix x0 fisher')
    
    B = axes[1,0].imshow(V_stretch_emp)
    pl.colorbar(B)
    axes[1,0].set_title('Covariance matrix stretch empirical')
    
    B = axes[1,1].imshow(V_stretch_fisher)
    pl.colorbar(B)
    axes[1,1].set_title('Covariance matrix stretch fisher')
    
    C = axes[2,0].imshow(V_color_emp)
    pl.colorbar(C)
    axes[2,0].set_title('Covariance matrix color empirical')
    
    C = axes[2,1].imshow(V_color_fisher)
    pl.colorbar(C)
    axes[2,1].set_title('Covariance matrix color fisher')
    
    #plot correlation matrix
    fig, axes = pl.subplots(figsize=(10,10), nrows=3, ncols=2)
    
    A = axes[0,0].imshow(corr_x0_emp)
    pl.colorbar(A)
    axes[0,0].set_title('Correlation matrix x0 empirical')
    
    A = axes[0,1].imshow(corr_x0_fisher)
    pl.colorbar(A)
    axes[0,1].set_title('Correlation matrix x0 fisher')
    
    B = axes[1,0].imshow(corr_stretch_emp)
    pl.colorbar(B)
    axes[1,0].set_title('Correlation matrix stretch empirical')
    
    B = axes[1,1].imshow(corr_stretch_fisher)
    pl.colorbar(B)
    axes[1,1].set_title('Correlation matrix stretch fisher')
    
    C = axes[2,0].imshow(corr_color_emp)
    pl.colorbar(C)
    axes[2,0].set_title('Correlation matrix color empirical')
    
    C = axes[2,1].imshow(corr_color_fisher)
    pl.colorbar(C)
    axes[2,1].set_title('Correlation matrix color fisher')



"""
This module is a rewrite of the minimizers.py module.
"""

import time
import logging

import numpy as np
import scipy
import scipy.optimize
import pandas
import matplotlib.pyplot as plt

from sksparse import cholmod

try:
    from sparse_dot_mkl import gram_matrix_mkl, dot_product_mkl
except:
    logging.warning('module: `sparse_dot_mkl` not available')
else:
    logging.info('sparse_dot_mkl found. Building hessian should be faster.')




class Minimizer:
    """Find the minimum of a function using a Newton-Raphson method
    """
    def __init__(self, log_likelihood, max_iter=100,
                 dchi2_stop=0.001, log=None):
        """
        """
        self.log_likelihood = log_likelihood
        # self.model = self.log_likelihood.model
        self.max_iter = max_iter
        self.dchi2_stop = dchi2_stop
        self._log = []

    def ndof(self):
        # return self.model.training_dataset.nb_meas() - len(self.model.pars.free)
        return self.log_likelihood.ndof()

    def get_log(self):
        return pandas.DataFrame(self._log)

    def _brent(self, pars, dpars):
        """attempt to save the day with a line search
        """
        def min_1d_func(t):
            return self.log_likelihood(pars.free + t * dpars.free)

        # ret = pars.copy()
        logging.info('linesearch (brent)...')
        t, val, ni, funcalls = \
            scipy.optimize.brent(min_1d_func, brack=(0., 1.),
                                 full_output=True)
        # ret.free = pars.free + t * dpars.free
        logging.info(f'done: t={t}, val={val}')
        # logging.info(f'{val} == {self.log_likelihood(pars.free)}')
        return t

    def minimize(self, p_init, mode='supernodal', ordering_method='metis',
                 beta=0., linesearch=False, dchi2_stop=None):
        """Minimize the log-likelihood
        """
        # pars = self.model.pars
        pars = self.log_likelihood.pars
        pars.free = p_init
        dchi2_stop = dchi2_stop if dchi2_stop is not None else self.dchi2_stop
        # dpars = self.model.pars.copy()
        dpars = pars.copy()
        old_pars = pars.copy()
        dpars.full[:] = 0.
        self._log = []

        # minimization loop
        old_chi2 = None
        for i in range(self.max_iter+1):
            logging.info(f'nacl.minimizer: {i}')
            chi2, grad, hessian = self.log_likelihood(pars.free,
                                                      deriv=True)
            try:
                logging.info('cholesky...')
                fact = cholmod.cholesky(hessian.tocsc(),
                                        mode=mode,
                                        ordering_method=ordering_method,
                                        beta=beta)
                logging.debug('done.')
            except:
                logging.error(f'cholesky failed: matrix non-posdef')
                # try:
                #         logging.warning('attempting to recover with LDLt')
                #         fact = cholmod.cholesky(hessian.tocsc(),
                #                                 mode='simplicial',
                #                                 ordering_method=ordering_method,
                #                                 beta=beta)
                # except:
                #     logging.error(f'failed also.')
                return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'Cholesky failed',
                        'grad': grad,
                        'hessian': hessian}
            dpars.free = fact(grad)

            # print(dpars.free)

            # store previous parameter values
            old_chi2 = chi2
            old_pars.free = pars.free
            pars.free = pars.free + dpars.free
            # and recompute the chi2
            chi2 = self.log_likelihood(pars.free, deriv=False)
            # chi2 decrement
            dchi2 = old_chi2 - chi2

            if dchi2 < 0 & (np.abs(dchi2) > dchi2_stop):
                logging.warning(f'increasing chi2: dchi2={dchi2:.4e}')
                t = self._brent(old_pars, dpars)
                pars.free = old_pars.free + t * dpars.free
                chi2 = self.log_likelihood(pars.free, deriv=False)
                dchi2 = old_chi2 - chi2
                if dchi2 < 0.:
                    logging.warning(f'increasing chi2: {old_chi2} -> {chi2}')
                    # revert to the previous step
                    pars.free = old_pars.free
                    return {'pars': pars.copy(),
                            'chi2': chi2,
                            'ndof': self.ndof(),
                            'status': 'increasing chi2'}

            # maybe we have converged ?
            if np.abs(dchi2) <= dchi2_stop:
                logging.info(f'converged: dchi2={dchi2:.4e}: {old_chi2:12.9g} -> {chi2:12.9g} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')
                return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'converged'}

            # and maybe we have exceeded the number of iterations ?
            if i >= self.max_iter:
                logging.info(f'iter {i:3} dchi2={dchi2:.4e}: {old_chi2:12.9g} -> {chi2:12.9g} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')
                return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'too many iterations'}

            logging.info(f'iter {i} dchi2={dchi2:.4e}: {old_chi2:12.9g} -> {chi2:12.9g} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')


    def minimize_lm(self, p_init, **kwargs):
        """The Levenberg Marquard version of minimize

        Parameters
        ----------
        dchi2_stop: float, default: 1.E-3
          stop criterion.
        mode: {'supernodal', 'simplicial'}, default: 'supernodal'
          cholmod algorithm.
        ordering_method: str, default: 'metis'
          ordering method.
        beta: float, default: 0.
          whether to surcharge the hessian diagonal before factorizing.
        lamb: float, default: 1.E-3
          the initial value of the LM damping parameter.
        accept: float, default: 10.
          factor applied on lambda if the step is accepted
        reject: float, default: 5.
          factor applied on lambda if the step is rejected
        max_iter: int, default:100
          maximum number of iterations before we give up
        geo: bool, default: False
          whether to enable geodesic acceleration
        hstep: float, default: 1.E-4
          numerical derivatives step
        Returns
        -------

        """
        # it has to be a copy !
        # pars = self.model.pars.copy()
        pars = self.log_likelihood.pars.copy()
        pars.free = p_init
        dchi2_stop = kwargs.get('dchi2_stop', self.dchi2_stop)
        # dpars = self.model.pars.copy()
        dpars = self.log_likelihood.pars.copy()
        dpars.full[:] = 0.
        self._log = []
        chi2_t = []
        tt = []
        
        old_pars = pars.copy()
        trial = pars.copy()
        mode = kwargs.get('mode', 'supernodal')
        ordering_method = kwargs.get('ordering_method', 'metis')
        beta = kwargs.get('beta', 0.)

        accept = kwargs.get('accept', 10.)
        reject = kwargs.get('reject', 10.)

        # number of tries to get a correct LM step
        max_attempts = kwargs.get('max_attempts', 100)
        # maximum number of iterations allowed
        max_iter = kwargs.get('max_iter', self.max_iter)
        lamb = kwargs.get('lamb', 10.)

        # diag_charge
        diag_charge = kwargs.get('diag_charge', 'levenberg')

        # whether to activate geodesic acceleration
        geo = kwargs.get('geo', False)
        h = kwargs.get('hstep', 1.E-4)

        # minimization loop
        for i in range(max_iter+1):
            # logging.info(f'minimize_lm: {i}')
            chi2, grad, hessian = self.log_likelihood(pars.free, deriv=True)
            # r0 = self.log_likelihood.w_res

            if i == 0:
                logging.info(f'init: l={lamb:6.1e} chi2: {chi2:12.4e} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')

            # LM step
            success, attempts = False, 0
            while not success:
                if attempts == max_attempts:
                    logging.error('unable to get a valid LM step')
                    # TODO: define what we should return here.
                    return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'no valid LM step'}

                try:
                    diag = hessian.diagonal()
                    if diag_charge == 'levenberg':
                        hessian.setdiag(diag + lamb)
                    elif diag_charge == 'marquardt':
                        hessian.setdiag((1.+lamb) * diag)
                    elif diag_charge == 'marquardt_max':
                        max_diag = diag.max()
                        hessian.setdiag(diag + lamb * max_diag)
                    else:
                        raise ValueError(f'diag_charge: invalid value {diag_charge}')

                    fact = cholmod.cholesky(hessian.tocsc(),
                                            mode=mode,
                                            ordering_method=ordering_method,
                                            beta=beta)

                    # NOTE: rectifying the algebra.
                    # if the log-likelohood returns its true gradient,
                    # the NR-step is H^-1 @ (-grad)
                    dpars.free = fact(-1. * grad)

                    # if geo:
                    #     _ = self.log_likelihood(pars.free + h * dpars.free)
                    #     r1 = self.log_likelihood.w_res
                    #     print(len(r1), len(r0))
                    #     d_w_res = (r1 - r0) / h
                    #     print(len(d_w_res), len(hessian.dot(dpars.free)))
                    #     dw = 2. * (d_w_res - hessian.dot(dpars.free)) / h
                    #     w_J = self.log_likelihood.w_J
                    #     geo_corr = -0.5 * fact(w_J.dot(dw))
                    #     truncerr = 2 * np.sqrt(geo_corr.dot(geo_corr)) / \
                    #                    np.sqrt(dpars.free.dot(dpars.free))

                    hessian.setdiag(diag)
                except:
                    logging.error(f'cholesky failed: matrix non posdef')
                    lamb *= reject
                    attempts += 1
                    continue

                # print(dpars)

                # old_chi2 = chi2
                # old_pars.free = pars.free
                # pars.free = pars.free + dpars.free
                trial.free = pars.free + dpars.free
                # if geo:
                #     trial.free += geo_corr

                # check_chi2 = self.log_likelihood(pars.free, deriv=False)
                trial_chi2 = self.log_likelihood(trial.free, deriv=False)
                # print(f'CHECK: {chi2} {check_chi2} {trial_chi2}')
                dchi2 = chi2 - trial_chi2

                # if geo and truncerr > 2.:
                #     pass
                if dchi2 > 0.  or (np.abs(dchi2) <= dchi2_stop):
                    success = True
                else:
                    pass

                # if dchi2 > 0. or (np.abs(dchi2) <= dchi2_stop):
                if success:
                    lamb /= accept
                    # logging.info(f'success: dchi2={dchi2}')
                else:
                    lamb *= reject
                    # chi2 = self.log_likelihood(pars.free, deriv=False)
                    logging.warning(f'[{attempts}/{max_attempts}] increasing chi2: dchi2={dchi2:.4e}')
                    logging.warning(f'next attempt with lambda={lamb:.4e}')
                    attempts += 1
                    continue

            # now, we can update the parameter vector
            old_chi2 = chi2
            chi2 = trial_chi2
            old_pars.free = pars.free
            pars.free = trial.free

            # if logging activated, we construct a log structure here
            l = self.log_likelihood.get_log()
            l['step'] = i
            l['time'] = time.perf_counter()
            l['lambda'] = lamb
            l['dchi2'] = dchi2
            l['ldpars'] = np.sqrt(np.dot(dpars.free, dpars.free))
            self._log.append(l)
            
            chi2_t.append(chi2)
            tt.append(time.perf_counter())
            
            # maybe we have converged ?
            if np.abs(dchi2) <= dchi2_stop:
                logging.info(f'converged: dchi2={dchi2:12.4e}: {old_chi2:12.4e} -> {chi2:12.4e} ndof={self.ndof()} chi2/ndof={chi2/self.ndof():.6f}')
                return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'converged',
                        'chi2_t':chi2_t,
                        'time':tt}

            # TODO: change this. This check should be after the loop
            # and maybe we have exceeded the number of iterations ?
            if i >= max_iter:
                logging.info(f'iter {i: 3d} l={lamb:6.1e}: {old_chi2:12.4e} -> {chi2:12.4e} | dchi2={dchi2:12.4e} | ndof={self.ndof()} chi2/ndof={chi2/self.ndof():.6f}')
                return {'pars': pars.copy(),
                        'chi2': chi2,
                        'ndof': self.ndof(),
                        'status': 'too many iterations',
                        'chi2_t':chi2_t,
                        'time':tt}

            logging.info(f'iter {i: 3d} l={lamb:6.1e}: {old_chi2:12.4e} -> {chi2:12.4e} | dchi2={dchi2:12.4e} | ndof={self.ndof()} chi2/ndof={chi2/self.ndof():.6f}')


    def get_cov_matrix(self, params_of_interest=None, corr=False, plot=False):
        """
        returns the covariance matrix and correlation matrix and the estimated
        errors on the parameters of interest
        cov : V
        corr : Vij/sqrt(Vii*Vjj)
        and can plot them
        
        Return:
        V : covariance matrix
        err_of_interest (optional) : errors on the specified parameters
        corr (optional) : correlation matrix
        """
        llk, grad, H = self.log_likelihood(self.log_likelihood.pars.free, deriv=True)
        #f = cholmod.cholesky(H, beta=1e-6)
        f = cholmod.cholesky(H)
        V = 2 * f.inv()
        var = V.diagonal()
        index_of_interest = np.array([])
        if plot:
            plt.figure()
            plt.imshow(V.toarray())
            plt.colorbar()
            plt.title('Covariance matrix')
        if params_of_interest is not None:
            for x in params_of_interest:
                idx = self.log_likelihood.pars.indexof(x)
                index_of_interest = np.append(index_of_interest, idx)
                if plot:
                    plt.figure()
                    plt.imshow(V.toarray()[idx][:,idx])
                    plt.colorbar()
                    plt.title('Covariance matrix of ' + x)
            index_of_interest = index_of_interest.astype(int)
            err_of_interest = np.sqrt(var[index_of_interest])
        else:
            err_of_interest = None
        if not corr:
            return V, err_of_interest
        else:
            N = len(var)
            i = np.arange(N)
            ii = np.vstack([i for k in range(N)])
            v1 = var[ii]
            v2 = v1.T
            vii_jj = v1 * v2
            vii_jj = np.sqrt(vii_jj)
            corr = V/vii_jj
            if plot:
                plt.figure()
                plt.imshow(corr.toarray())
                plt.colorbar()
                plt.title('Correlation matrix')
            return V, err_of_interest, corr
            
        
    def plot(self, timesteps=False):
        """
        """
        if not hasattr(self, '_log'):
            return

        import pylab as pl

        d = pandas.DataFrame(self._log)
        fig, axes = plt.subplots(nrows=6, ncols=1, figsize=(6,12), sharex=True)
        xx = d.step if not timesteps else d.time-d.time.min()
        xlabel = 'step' if not timesteps else 'time [s]'
        axes[0].plot(xx, d.llk, 'k.:')
        axes[0].set_ylabel('$-2\ln {\cal L}$')
        axes[1].semilogy(xx, d.main_chi2, 'b.:')
        axes[1].set_ylabel('$\chi^2$')
        axes[2].plot(xx, d.log_det_v, 'r.:')
        axes[2].set_ylabel('$\log\|\mathbf{V}\|$')
        axes[3].semilogy(xx, d.reg, 'g.:')
        axes[3].set_ylabel('reg')
        axes[4].semilogy(xx, d.cons, 'g.:')
        axes[4].set_ylabel('cons')
        axes[5].semilogy(xx, d.dchi2, 'b.:')
        axes[5].set_ylabel('$\delta -2\ln {\cal L}$')
        axes[5].set_xlabel(xlabel)
        plt.subplots_adjust(wspace=0.005, hspace=0.005)

    # this version will be deprecated in the future
    def __call__(self, p_init, mode='supernodal', ordering_method='metis',
                 beta=0., linesearch=False):
        """Minimize the log-likelihood
        """
        # pars = self.model.pars
        pars = self.log_likelihood.pars
        pars.free = p_init
        # dpars = self.model.pars.copy()
        dpars = self.log_likelihood.pars.copy()
        pars_before = pars.copy()
        dpars.full[:] = 0.
        self._log = []

        # minimization loop
        old_chi2 = None
        for i in range(self.max_iter):
            logging.info(f'nacl.minimizer: {i}')
            chi2, grad, hessian = self.log_likelihood(pars.free,
                                                      deriv=True)
            # maybe we are close to the minimum already ?
            if old_chi2 is not None:
                dchi2 = old_chi2 - chi2
                if (np.abs(dchi2) < self.dchi2_stop):
                    logging.info(f'converged: dchi2={dchi2:.4e}: {old_chi2:12.9g} -> {chi2:12.9g} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')
                    return {'pars': pars.copy(),
                            'chi2': chi2,
                            'ndof': self.ndof(),
                            'status': 'converged'}
                if dchi2 < 0.:
                    logging.warning(f'increasing chi2: {old_chi2} -> {chi2}')
                    # revert to the previous step
                    pars.free = pars_before.free
                    return {'pars': pars.copy(),
                            'chi2': chi2,
                            'ndof': self.ndof(),
                            'status': 'increasing chi2'}
                if i >= self.max_iter:
                    return {'pars': pars.copy(),
                            'chi2': chi2,
                            'ndof': self.ndof(),
                            'status': 'too many iterations'}
                logging.info(f'iter {i} dchi2={dchi2:.4e}: {old_chi2:12.9g} -> {chi2:12.9g} ndof={self.ndof()} chi2/ndof={chi2/self.ndof()}')
            # else:
                # logging.info(f'first step - ')
            # if not, let's compute the next Newton-Raphson step...
            try:
                logging.info('cholesky...')
                fact = cholmod.cholesky(hessian.tocsc(),
                                        mode=mode,
                                        ordering_method=ordering_method,
                                        beta=beta)
                logging.info('done.')
            except:
                logging.error(f'cholesky failed: matrix non-posdef')
                return {'pars': pars.copy,
                        'chi2': chi2,
                        'ndof': len(pars.free),
                        'status': 'Cholesky failed',
                        'grad': grad,
                        'hessian': hessian}
            dpars.free = fact(grad)

            if linesearch:

                def min_1d_func(t):
                    return self.log_likelihood(pars_before.free + \
                                               t * dpars.free)

                logging.info('linesearch (brent)...')
                t, val, ni, funcalls = \
                    scipy.optimize.brent(min_1d_func, brack=(0., 1.),
                                         full_output=True)
                pars.free = pars_before.free + t * dpars.free
                logging.info(f'done: t={t}, val={val}')
                logging.info(f'{val} == {self.log_likelihood(pars.free)}')

            else:
                pars.free = pars_before.free + dpars.free

            if old_chi2 and (old_chi2 < chi2):
                # if increasing chi2 detected, then linesearch
                t = self._brent(pars_before, dpars)
                pars.free = pars_before.free + dpars.free

            # and update the old values
            old_chi2 = chi2
            pars_before.free = pars.free


# class LogLikelihood2:
#     """Compute a LogLikelihood and its derivatives from a model.

#     The main purpose of this class, is to assemble the ingredients of the
#     linearized normal equations, from (1) a model (2) a dataset (3) optional
#     add-ons such as contraints, regularization and an error model.

#     It was written to mimimize the typical Likelihood of the NaCl model
#     (see Guy's thesis, equation (1) page XX).
#     """

#     def __init__(self, model, cons=None, reg=None, variance_model=None,
#                  spec_weight_scale=1.,
#                  phot_error_pedestal=None,
#                  force_default_spgemm=False):
#         """
#         """
#         self.model = model
#         self.training_dataset = model.training_dataset
#         self.pars = self.model.pars.copy()
#         self.cons = cons if cons is not None else []
#         self.reg = reg if reg is not None else []
#         self.variance_model = variance_model
#         self.spec_weight_scale = spec_weight_scale
#         self.y = self.training_dataset.get_all_fluxes()
#         self.yerr = self.training_dataset.get_all_fluxerr()
#         self.bads = self.training_dataset.get_valid() == 0
#         self.phot_error_pedestal = phot_error_pedestal
#         self.force_default_spgemm = force_default_spgemm
#         self._log = None

#     def get_log(self):
#         return self._log

#     def ndof(self):
#         """
#         """
#         nphot, nsp, nspp = self.training_dataset.nb_meas(valid_only=True,
#                                                          split_by_type=True)
#         nmeas = nphot + np.sqrt(self.spec_weight_scale) * nsp + nspp
#         n_free_pars = len(self.model.pars.free)
#         return int(nmeas - n_free_pars)

#     def __call__(self, p, deriv=False):
#         """evaluate for the current parameters
#         """
#         self.model.pars.free = p

#         # the ingredients
#         bads = self.bads
#         model_flux, model_jac = None, None
#         model_var, model_var_jac = 0., None
#         chi2, log_det_v = 0., 0.

#         # just for logging purposes
#         main_chi2 = 0.
#         cons_chi2 = []
#         reg_chi2 = []

#         # wscales
#         # WARNING : in this method, the weights correspond to the
#         # inverse of the sigma, and not the inverse of the variance
#         # the weight scales follow this convention. Hence, to
#         # reduce the contribution of the spectra to the global chi2
#         # by a factor 0.1, the wscales need to be equal to np.sqrt(0.1)
#         wscales = np.ones(len(self.y))
#         nlc, nsp, nspp = self.training_dataset.nb_meas(valid_only=False, split_by_type=True)
#         wscales[nlc:nlc+nsp] *= np.sqrt(self.spec_weight_scale)

#         # if no derivatives requested,
#         # just add the penalities and we should be ok
#         if not deriv:
#             # model and variance
#             model_flux = self.model(p, jac=False)
#             var = self.yerr**2 / wscales ** 2
#             # here: add an optional pedestal to fluxerr
#             # warning: this option is dangerous, because
#             # it results in the biased result. Use it only
#             # during the first steps of training, to ease
#             # the first chi2 minimizations
#             if self.phot_error_pedestal is not None:
#                 nlc, _, _ = self.training_dataset.nb_meas(valid_only=False, split_by_type=True)
#                 ped = (self.phot_error_pedestal * self.y) ** 2
#                 var[:nlc] += (ped[:nlc] / wscales**2)
#             if self.variance_model is not None:
#                 model_var = self.variance_model(model_flux=model_flux)
#                 var += (model_var / wscales**2)
#                 log_det_v = np.log(var[~bads]).sum()

#             # weighted residuals
#             res = self.y - model_flux
#             sig = np.sqrt(var)
#             wres = res / sig

#             # if hasattr(self, 'debug'):
#             #     self.chi2_debug = (wres[~bads]**2).sum()
#             #     self.log_det_v_debug = log_det_v
#             #     self.full_chi2_debug = self.chi2_debug + log_det_v

#             # and the chi2
#             main_chi2 = (wres[~bads]**2).sum()
#             chi2 = main_chi2 + log_det_v
#             for cons in self.cons:
#                 c = cons(p, deriv=False)
#                 cons_chi2.append(c)
#                 chi2 += c
#                 # chi2 += cons(p, deriv=False)
#             for reg in self.reg:
#                 r = reg(p, deriv=False)
#                 reg_chi2.append(r)
#                 chi2 += r
#                 # chi2 += reg(p, deriv=False)

#             self._log = {'chi2': main_chi2,
#                          'log_det_v': log_det_v,
#                          'cons': np.sum(cons_chi2),
#                          'reg': np.sum(reg_chi2)}

#             return chi2

#         # that was the easy part. Now, when derivatives requested
#         # things are little more complicated.
#         logging.debug('model (jac=True)')
#         model_flux, model_jac = self.model(p, jac=True)
#         var = self.yerr**2 / wscales**2
#         if self.phot_error_pedestal is not None:
#             nlc, _, _ = self.training_dataset.nb_meas(valid_only=False, split_by_type=True)
#             ped = (self.phot_error_pedestal * self.y) ** 2
#             var[:nlc] += (ped[:nlc] / wscales**2)
#         if self.variance_model is not None:
#             logging.debug('variance model (jac=True)')
#             model_var, model_var_jac = \
#                 self.variance_model(model_flux=model_flux,
#                                     model_jac=model_jac,
#                                     jac=True)
#             var += (model_var / wscales**2)
#             # I assume model_var_jac is a coo_matrix
#             model_var_jac = model_var_jac.tocoo()
#             model_var_jac.data /= wscales[model_var_jac.row]**2
#             log_det_v = np.log(var[~bads]).sum()

#         # weighted residuals
#         res = self.y - model_flux
#         w = 1. / np.sqrt(var)
#         N = len(self.y)
#         W = scipy.sparse.dia_matrix((w, [0]), shape=(N, N))
#         w_res = W @ res
#         w_J = W @ model_jac

#         # cut the bads
#         w_J = w_J[~bads,:]
#         w_res = w_res[~bads]

#         # store the results. May be used later
#         self.w_res = w_res
#         self.w_J = w_J

#         # chi2
#         main_chi2 = chi2 = (w_res**2).sum()
#         chi2 += log_det_v

#         # if hasattr(self, 'debug'):
#         #     self.chi2_debug = main_chi2
#         #     self.log_det_v_debug = log_det_v
#         #     self.full_chi2_debug = chi2

#         # the gradient and hessian have several components
#         # first, the two classical ones: J^TWJ and J^TWR
#         grad = 2. * w_J.T @ w_res
#         w_J = w_J.tocsr()
#         if 'gram_matrix_mkl' in globals() and not self.force_default_spgemm:
#             logging.debug('hessian: H = J.T @ J (gram_matrix_mkl)')
#             if 1 in w_J.shape:
#                 # there seem to be a bug in gram_matrix_mkl when
#                 # one of the matrix dimensions is one. typically refuses
#                 # to contract (N,1) matrix into a scalar
#                 hess = 2. * dot_product_mkl(w_J.T, w_J, reorder_output=True)
#             else:
#                 # `reorder_output=True` seems essential here.
#                 # otherwise, the Newton-Raphson step is wrong, typically
#                 # by a factor 0.5 ... this is scary, I know...
#                 hess = 2. * gram_matrix_mkl(w_J, reorder_output=True)
#                 row, col = hess.nonzero()
#                 hess[col,row] = hess[row,col]
#                 #            hess = 2. * dot_product_mkl(w_J.T, w_J, reorder_output=True)
#         else:
#             logging.debug('hessian: H = J.T @ J (slow version)')
#             hess = 2. * w_J.T @ w_J

#         # model_jac.data *= w[model_jac.row]
#         # main_chi2 = chi2 = (wres[~bads]**2).sum()
#         # chi2 += log_det_v
#         # wres = wres[~bads]

#         # then, the contributions of the (optional) variance model
#         # to the hessian and gradient
#         # TODO:
#         #  - remove outliers from the model jacobian
#         #  - if W @ J faster, do that
#         # TODO:
#         #  -if some variance model parameters are fixed this is apparently
#         # not taken into account ...
#         if self.variance_model is not None:
#             logging.debug('variance model')
#             WW = scipy.sparse.dia_matrix((w**2, [0]), shape=(N, N))
#             # model_var_jac.data /= var[model_var_jac.row]
#             model_var_jac = (WW @ model_var_jac)[~bads, :]

#             # gradient
#             # was +1
#             # rWdVWr = -1. * (w_res**2).sum() * np.array(model_var_jac.T.sum(axis=1)).squeeze()
#             mvJ = model_var_jac.tocoo()
#             wres_mvJ = scipy.sparse.coo_matrix(
#                 (mvJ.data * w_res[mvJ.row],
#                  (mvJ.row, mvJ.col)),
#                  shape=mvJ.shape)
#             rWdVWr = 1. * wres_mvJ.T @ w_res

#             # rWdVWr = -1. * (w_res**2 * np.array(model_var_jac.T)).sum(axis=1).squeeze()
#             grad += rWdVWr
#             # was -1
#             tr_WdV = -1. * np.array(model_var_jac.T.sum(axis=1)).squeeze()
#             grad += tr_WdV

#             # hessian
#             tr_WdVWdV = 1. * model_var_jac.T.dot(model_var_jac)
#             hess += tr_WdVWdV

#         # the quadratic constraints
#         logging.debug('constraints')
#         for penality in self.cons:
#             v_pen, grad_pen, hess_pen = penality(p, deriv=True)
#             chi2 += v_pen
#             grad += grad_pen
#             hess += hess_pen
#             cons_chi2.append(v_pen)

#         # the regularization
#         logging.debug('regularization')
#         for penality in self.reg:
#             v_pen, grad_pen, hess_pen = penality(p, deriv=True)
#             chi2 += v_pen
#             grad += grad_pen
#             hess += hess_pen
#             reg_chi2.append(v_pen)

#         self._log = {
#             'chi2': main_chi2,
#             'log_det_v': log_det_v,
#             'cons': np.sum(cons_chi2),
#             'reg': np.sum(reg_chi2),
#         }

#         msg = f'chi2={main_chi2:.6e} | log_det_v={log_det_v} | cons='
#         for cons_val in cons_chi2:
#             msg += f'{cons_val:8.6e}'
#         msg += ' | reg='
#         for reg_val in reg_chi2:
#             msg += f'{reg_val:.6e}'
#         logging.debug(msg)

#         return chi2, grad, hess

#     def get_weighted_residuals(self, p):
#         """a utility function to get the weighted residuals
#         """
#         model_flux = self.model(p, jac=False)
#         res = self.y - model_flux
#         var = self.yerr**2
#         if self.variance_model is not None:
#             model_var = self.variance_model(model_flux=model_flux,
#                                             jac=False)
#             var += model_var
#         w = 1. / np.sqrt(var)
#         W = scipy.sparse.dia_matrix((w, [0]), shape=(N,N))
#         w_res = W @ res

#         return w_res


# class LogLikelihood:
#     """Compute a LogLikelihood and its derivatives from a model.

#     The main purpose of this class, is to assemble the ingredients of the
#     linearized normal equations, from (1) a model (2) a dataset (3) optional
#     add-ons such as contraints, regularization and an error model.

#     It was written to mimimize the typical Likelihood of the NaCl model
#     (see Guy's thesis, equation (1) page XX).
#     """

#     def __init__(self, wres_func, cons=None, reg=None, error_snake=None):
#         """
#         """
#         self.func = wres_func
#         self.model = self.func.model
#         self.pars = self.func.model.pars.copy()
#         self.cons = cons if cons is not None else []
#         self.reg = reg if reg is not None else []
#         self.error_snake = error_snake if error_snake is not None else []
#         self.bads = wres_func.model.training_dataset.get_valid() == 0

#     def __call__(self, p, deriv=False):
#         """evaluate for the current parameters
#         """
#         self.model.pars.free = p
#         bads = self.bads

#         # if no derivatives requested,
#         # just add the penalities and we should be ok
#         if not deriv:
#             w_res = self.func(p, jac=False)
#             chi2 = (w_res[~bads]**2).sum()
#             for cons in self.cons:
#                 chi2 += cons(p, deriv=False)
#             for reg in self.reg:
#                 chi2 += reg(p, deriv=False)
#             return chi2

#         # if derivatives requested, then
#         # 1) base chi2 + gradient + hessian
#         w_res, w_J, _ = self.func(p, jac=True)
#         w_res = w_res[~bads]
#         w_J = w_J[~bads,:]
#         chi2 = (w_res**2).sum()
#         grad = 2. * w_J.T @ w_res
#         hess = 2. * w_J.T @ w_J

#         # just for logging purposes
#         main_chi2 = chi2
#         cons_chi2 = []
#         reg_chi2 = []

#         # add the penalities and their derivatives
#         for penality in self.cons:
#             v_pen, grad_pen, hess_pen = penality(p, deriv=True)
#             chi2 += v_pen
#             grad += grad_pen
#             hess += hess_pen
#             cons_chi2.append(v_pen)

#         for penality in self.reg:
#             v_pen, grad_pen, hess_pen = penality(p, deriv=True)
#             chi2 += v_pen
#             grad += grad_pen
#             hess += hess_pen
#             reg_chi2.append(v_pen)

#         msg = f' LogLikelihood: chi2={main_chi2:.6e} | cons='
#         for cons_val in cons_chi2:
#             msg += f'{cons_val:.6e}'
#         msg += ' | reg='
#         for reg_val in reg_chi2:
#             msg += f'{reg_val:.6e}'
#         logging.info(msg)

#         return chi2, grad, hess


class WeightedResiduals:

    def __init__(self, model, variance_model=None):
        """constructor - iniate the residual function
        """
        self.model = model
        self.training_dataset = model.training_dataset
        self.variance_model = variance_model
        self.y = self.training_dataset.get_all_fluxes()
        self.yerr = self.training_dataset.get_all_fluxerr()

    def __call__(self, p, jac=False):
        """evaluate
        """
        self.model.pars.free = p

        # evaluate the residuals
        val, J = None, None
        if jac:
            val, J = self.model(p, jac=True)
        else:
            val = self.model(p, jac=False)
        res = self.y - val

        # measurement variance and weights
        var, Jvar = 0., None
        if self.variance_model is not None:
            if not jac:
                var = self.variance_model(p, jac=False)
            else:
                var, Jvar = self.variance_model(p, jac=True)

        res_var = self.yerr**2 + var
        w = 1. / np.sqrt(res_var)
        N = len(self.y)
        W = scipy.sparse.dia_matrix((w, [0]), shape=(N, N))

        # weighted residuals
        wres = W @ res
        if jac:
            wJ = W @ J
            return wres, wJ, Jvar
        return wres



import os
import os.path as op
import numpy as np
import scipy.sparse
import saunerie
import croaks
from saunerie.fitparameters import FitParameters

from salt3.minimizers import minimize_nr_var_quad,minimize_nr
import salt3.generator
import salt3.models
import salt3.plots

from helpers import locate_test_datafile
from helpers import check_grad


class ToySplineModelFull():
    
    def __init__(self, dp, phase_range=[-100., 100.], grid_size=30):
        self.dp = dp
        self.nb_sn = len(dp.sn_set)
        self.spline_basis = saunerie.bspline.CardinalBSplineC(grid_size, phase_range)
        self.n_theta = grid_size
        self.pars = self.init_pars(dp)

        sn_index = dp.sn_index
        flx = self.pars['flux'].full[sn_index]
        tmax = self.pars['tmax'].full[sn_index]
        s = self.pars['s'].full[sn_index]   
        phases = (self.dp.date-tmax) / (1.+s)
        self.jacobian = self.spline_basis.eval(phases)
                
    def init_pars(self, dp):
        fp = FitParameters([('flux', self.nb_sn),
                            ('tmax', self.nb_sn),
                            ('s', self.nb_sn),
                            ("theta", len(self.spline_basis))])
        fp['flux'].full[:] = 1.
        fp['tmax'].full[:] = 0.
        fp['s'].full[:] = 0.
        fp['theta'].full[:] = 1.E-6
        return fp

    def __call__(self, p, jac=False, dp=None):
        """
        """
        self.pars.free = p

        if dp is None:
            dp = self.dp

        sn_index = dp.sn_index
        flx = self.pars['flux'].full[sn_index]
        tmax = self.pars['tmax'].full[sn_index]
        s = self.pars['s'].full[sn_index]        
        phases = (dp.date-tmax)  / (1.+s)
        
#        if dp is not None:
#            jacobian = self.spline_basis.eval(phases)
#        else:
#            jacobian = self.jacobian

        jacobian = self.spline_basis.eval(phases)

        v = flx * (jacobian @ self.pars['theta'].full)

        if jac:
            N = len(dp.nt)
            n = len(self.spline_basis)
            i = np.arange(N)

            # where are phases defined ?
            dB = self.spline_basis.deriv(phases)
            
            # dmodel_dflux 
            i_dm_dflux = i
            j_dm_dflux = self.pars['flux'].indexof(sn_index)
            v_dm_dflux = jacobian @ self.pars['theta'].full

            # dmodel_dtmax
            i_dm_dtmax = i
            j_dm_dtmax = self.pars['tmax'].indexof(sn_index)
            v_dm_dtmax = flx * (-1. / (1.+s)) * (dB @ self.pars['theta'].full)

            # dmodel_ds
            i_dm_ds = i
            j_dm_ds = self.pars['s'].indexof(sn_index)
            v_dm_ds = flx * (-(dp.date-tmax)/(1+s)**2) * (dB @ self.pars['theta'].full)

            # dmodel_dtheta
            i_dm_dtheta = jacobian.row
            j_dm_dtheta = self.pars['theta'].indexof(jacobian.col)
            v_dm_dtheta = self.pars['flux'].full[sn_index[jacobian.row]] * jacobian.data
            
            i = np.hstack((i_dm_dflux, i_dm_dtmax, i_dm_ds, i_dm_dtheta))
            j = np.hstack((j_dm_dflux, j_dm_dtmax, j_dm_ds, j_dm_dtheta))
            val = np.hstack((v_dm_dflux, v_dm_dtmax, v_dm_ds, v_dm_dtheta))
            
            idx_fix = np.where(j != -1)[0]
            DZ = scipy.sparse.coo_matrix((val[idx_fix], (i[idx_fix], j[idx_fix])), shape=(N, len(self.pars.free)))
            #            DZ.data /=  -dp.flux_err[DZ.row]
            
            # return (dp.flux-v)/flux_err, DZ
            return v, DZ
        
        return v

class ToySplineRegularization():
    
    def __init__(self, pars, mu=1.E-8):
        self.pars = pars
        self.mu = mu

        # build the regularization matrix
        n = len(self.pars['theta'].full)
        data = np.ones(n).astype(float)
        M = self.mu * scipy.sparse.dia_matrix((data, [0]), shape=(n,n)).tocoo()

        # build the large matrix which corresponds to the model parameters
        pp = pars.copy()
        pp.release()
        N = len(pp.full)
        i = pp['theta'].indexof(M.row)
        j = pp['theta'].indexof(M.col)
        idx = (i>=0) & (j>=0) # should be all true's
        self.M = scipy.sparse.coo_matrix((M.data[idx], (i[idx],j[idx])), shape=(N,N)).tocsr()
        

    def __call__(self, p, jac=True):
        self.pars.free = p
        v = float(self.pars.full.T @ self.M @ self.pars.full)
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            # there might be a more efficient way to slice the matrix ...
            return v, self.M[i,:][:,i]
        return v
    

class ToySplineNonIntegralConstraints():
    """
    """
    def __init__(self, pars, basis, vals=[1., 0.]):
        self.pars = pars
        self.basis = basis
        self.vals = self.d = np.array(vals)
        assert(len(self.d) == 2)
        
        n = len(self.pars.full)
        
        # first constraint: S(0) = 1
        jacobian = self.basis.eval(np.array([0]))
        i1 = pars['theta'].indexof(jacobian.col)
        j1 = np.full(len(jacobian.col), 0)
        v1 = jacobian.data.copy()
        # second constraint: S'(0) = 0
        Jp = self.basis.deriv(np.array([0]))
        i2 = pars['theta'].indexof(Jp.col)
        j2 = np.full(len(Jp.col), 1)
        v2 = Jp.data.copy()
        self.H = scipy.sparse.coo_matrix((np.hstack((v1,v2)), (np.hstack((i1,i2)), np.hstack((j1,j2)))), shape=(n,2)).tocsr()
        
    #    def __len__(self):
    #        return 2

    def get_rhs(self):
        return self.d
        
    def __call__(self, p, jac=False):
        self.pars.free = p
        v = -self.H.T @ self.pars.full + self.vals
        
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            return v, self.H[i,:].T
        return v

    
class ToySplineNonIntegralConstraintsFull():
    """
    .. note:: what if one of the constraints is invalid 
       because all corresponding parameters have been fixed ? 
    """
    def __init__(self, pars, basis, vals=[1., 0., 0.]):
        self.pars = pars
        self.basis = basis
        self.vals = self.d = np.array(vals)
        assert(len(self.d) == 3)
        
        n = len(self.pars.full)

        pp = pars.copy()
        pp.release()
        nsn = len(pp['flux'].full)
        
        # first constraint: S(0) = 1
        jacobian = self.basis.eval(np.array([0]))
        #        i1 = pars['theta'].indexof(jacobian.col)
        i1 = pp['theta'].indexof(jacobian.col)
        j1 = np.full(len(jacobian.col), 0)
        v1 = J.data.copy()
        # second constraint: S'(0) = 0
        Jp = self.basis.deriv(np.array([0]))
        #        i2 = pars['theta'].indexof(Jp.col)
        i2 = pp['theta'].indexof(Jp.col)        
        j2 = np.full(len(Jp.col), 1)
        v2 = Jp.data.copy()
        # third constraint: \sum_i s_i = 0
        i3 = pp['s'].indexof(np.arange(nsn))
        j3 = np.full(nsn, 2)
        v3 = np.full(nsn, 1)
        
        self.H = scipy.sparse.coo_matrix((np.hstack((v1,v2,v3)), (np.hstack((i1,i2,i3)), np.hstack((j1,j2,j3)))), shape=(n,3)).tocsr()
        
    #    def __len__(self):
    #        return 3

    def get_rhs(self):
        return self.d
    
    def __call__(self, p, jac=False):
        self.pars.free = p
        v = -self.H.T @ self.pars.full + self.vals
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            return v, self.H[i,:].T
        return v
    

def main(N=200, minimize=minimize_nr,#_var_quad,
         mu_regularization=500.,
         plot=False, seed=None):
    """
    """
    lc_data = salt3.generator.generate_lc(N, tmax_range=(0., 0.),
                                          sig_range=(0., 0.),
                                          norm_range=(1000., 1000.),
                                          seed=seed)

    model = ToySplineModelFull(lc_data, grid_size=250) # phase_range=[lc_data.date.min(), lc_data.date.max()], grid_size=100)
    f = salt3.models.VarModelResiduals(lc_data, model)
    reg = ToySplineRegularization(f.pars, mu=mu_regularization)
    cons = ToySplineNonIntegralConstraints(model.pars, model.spline_basis)
    cons_full = ToySplineNonIntegralConstraintsFull(model.pars, model.spline_basis)

    # initialization fit (spline parameters only)
    f.pars['flux'].full[:] = 1000
    f.pars['theta'].full[:] = 1.E-6
    f.pars['flux'].fix()
    f.pars['tmax'].fix()
    f.pars['s'].fix()
    x = minimize(f, f.pars.free, reg=reg, cons=cons)
    if len(x) == 2:
        x = x[0]
    f.pars.free = x

    # real fit, with all parameters free
    f.pars.release()
    f.pars['flux'].full[:] = 10000
    x = minimize(f, f.pars.free, cons=cons_full, reg=reg)
    if len(x) == 2:
        x = x[0]
    f.pars.free = x
    
    if plot:
        import pylab as pl
        pl.ion()
        import matplotlib as mpl
        pl.figure()
        gs = mpl.gridspec.GridSpec(4,1)
        ax = pl.subplot(gs[0:3,0])
        pl.errorbar(lc_data.date, lc_data.flux, yerr=lc_data.flux_err, ls='', marker='.', color='b')
        xx = np.linspace(lc_data.date.min(), lc_data.date.max(), 100)
        xx.sort()

        try:
            tmax = f.pars['tmax'].full
        except:
            tmax = np.zeros(N)
        xdata = salt3.plots.get_xdata_for_model_evaluation(N, tmax, lc_data.band_set, npoints=200)
        v = model(f.pars.free, dp=xdata, jac=False)
        for sn in lc_data.sn_set:
            idx = xdata.sn == sn
            pl.plot(xdata.date[idx], v[idx], ls='-', color=pl.cm.jet(sn/xdata.sn.max()))
        
        pl.subplot(gs[3,0], sharex=ax)
        pl.errorbar(lc_data.date, f(x,jac=False)*lc_data.flux_err, yerr=lc_data.flux_err, marker='.', ls='', color='b')
        pl.grid(1)
        pl.xlabel('x')
        pl.ylabel('res')
        pl.subplots_adjust(hspace=0.05)        

        pl.figure()
        xx = np.linspace(-50., 50., 10000)
        J = model.spline_basis.eval(xx)
        pl.plot(xx, J @ model.pars['theta'].full, 'r-')
        pl.xlabel('x')
        pl.ylabel('LC template (spline model only)')
        pl.grid(1)


        
    return lc_data, f, cons_full, reg, x


def test_model_derivatives(N=200, seed=0):
    """
    """
    lc_data = salt3.generator.generate_lc(N=N, tmax_range=(-70., 70.),
                                          sig_range=(-1., 1.),
                                          norm_range=(10., 1000.),
                                          seed=seed)

    model = ToySplineModelFull(lc_data, grid_size=250) # phase_range=[lc_data.date.min(), lc_data.date.max()], grid_size=100)
    f = salt3.models.ModelResiduals(lc_data, model)
    
    J, JJ = check_grad(model, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-8)
    
    J, JJ = check_grad(f, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model.pars.free[:] += np.random.uniform(-1., 1., len(model.pars.free))
    J, JJ = check_grad(model, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    J,JJ = check_grad(f, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    # en revanche, celui la, il plante ... 
    #    model.pars.full[:] += np.random.uniform(-1., 1., len(model.pars.full))
    #    J, JJ = check_grad(model, model.pars.free)
    #    D = np.abs(J-JJ)
    #    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    #    J,JJ = check_grad(f, model.pars.free)
    #    D = np.abs(J-JJ)
    #    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    
    
def test_1dsplinemodelfullfit_nr():
    x_target = np.load(locate_test_datafile('test_1dsplinemodel_solution.npy'))
    lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr,#_var_quad,
                               mu_regularization=10,
                               plot=0, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)    
    

if __name__ == "__main__":
    lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr,#_var_quad,
                               mu_regularization=10., plot=True)
    



"""test of the new bandpass system (a la sncosmo)
"""

import numpy as np
from nacl.lib.bspline import BSpline
import nacl.bandpasses



def test_filterdb_insertion():
    """
    """
    grid = np.arange(2000., 11010., 10.)
    basis = BSpline(grid)
    db = nacl.bandpasses.FilterDb(basis)

    g = nacl.bandpasses.get_bandpass('megacampsf::g', 12.)
    db.insert(g)
    db.insert(g, 0.325)
    r = nacl.bandpasses.get_bandpass('megacampsf::r', 1.233)
    db.insert(r, 0.766)
    db.insert(r)

    assert len(db) == 4
    a = db[(g.name, 0.325)]
    b = db[g.name]
    c = db[r.name]


def test_filterdb_pickle():
    """
    """
    grid = np.arange(2000., 11010., 10.)
    basis = BSpline(grid)
    db = nacl.bandpasses.FilterDb(basis)

    g = nacl.bandpasses.get_bandpass('megacampsf::g', 12.)
    db.insert(g)
    db.insert(g, 0.325)
    r = nacl.bandpasses.get_bandpass('megacampsf::r', 1.233)
    db.insert(r, 0.766)
    db.insert(r)

    import pickle
    with open('fdb.pkl', 'wb') as f:
        pickle.dump(db, f)

    with open('fdb.pkl', 'rb') as f:
        ddb = pickle.load(f)

    # compare the contents of the two databases
    # TODO
    



import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

import numpy as np
import pylab as pl
pl.ion()
from nacl.dataset import TrainingDataset
from nacl.train import TrainSALT2Like
from nacl.models import salt2
import pandas as pd
import os
os.environ['KMP_DUPLICATE_LIB_OK']='True'
from lemaitre import bandpasses

if __name__ == '__main__':
    fl = bandpasses.get_filterlib()
    tds = TrainingDataset.read_parquet('../data/lemaitre_full/lemaitre.compressed.parquet', filterlib=fl)
    
    #KEEPING ONLY HALF OF ZTF SNe
    """idx = tds.sn_data.nt['survey'] == 'ZTF'
    sn_remove = tds.sn_data.nt[idx]['name'][1500:]
    tds.kill_sne(sn_remove)
    tds.compress()
    
    
    #trainer = TrainSALT2Like(tds, variance_model='simple_snake')
    trainer = TrainSALT2Like(tds, variance_model='sn_local_snake')
    trainer.train_salt2_model()
    """
    
    #KEEPING ONLY HALF OF ZTF SNe
    # idx = tds.sn_data.nt['survey'] == 'ZTF'
    # sn_remove = tds.sn_data.nt[idx]['name'][:1500]
    # tds.kill_sne(sn_remove)
    # tds.compress()
    
    
    #trainer = TrainSALT2Like(tds, variance_model='simple_snake')
    trainer = TrainSALT2Like(tds, variance_model='sn_local_snake')
    trainer.train_salt2_model()



"""
"""

import numpy as np
# from scipy import sparse
# from sksparse import cholmod

import pandas
from ....lib.dataproxy import DataProxy
from ....lib.bspline import BSpline



def gaussian(x, sigma):
    """
    Gaussian function *2
    """
    # norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    return np.exp(-0.5 * xx**2)


class ToyModelGenerator:

    def __init__(self, nsn=100, npts=10, nbands=1,
                 xmin=-10., xmax=10., delta=10.,
                 yerr=0.01,
                 error_pedestal=None, error_snake=None,
                 wl_pivot=5000.,
                 v_calib=0.005**2, color_scatter=np.array([0.02, 0., 0.002])):
        """
        .. note::
           error snake and error_pedestal are mutually exclusive.
           if error_snake > 0, then error_pedestal is set to zero
        """
        self.nsn = nsn
        self.npts = npts
        self.nbands = nbands
        self.xmin = xmin
        self.xmax = xmax
        self.delta = delta
        self.yerr = yerr
        self.error_pedestal = 0.
        self.error_snake = 0.
        if error_pedestal is not None:
            self.error_pedestal = error_pedestal
        if error_snake is not None and error_snake > 0.:
            self.error_pedestal = 0.
            self.error_snake = error_snake
        self.wl_pivot = wl_pivot
        self.band_wl = 1500 * np.arange(nbands) + 4000.
        self.v_calib = v_calib
        if type(v_calib) is float:
            self.v_calib = np.diag(np.full(self.nbands, v_calib))
        self.color_scatter = color_scatter

        self.sn_pars = {'x0_range': (0.1, 100.), 'tmax_margin': 5., 'stretch_range': (0., 0.25), 'color_pars': (0., 0.1)}
        self.lcpars = {'g': (0., 0.8), 'r': (0.2, 1.2), 'i': (0.7, 2.)}

    def gen_sample(self):
        """
        """
        # values and errors
        if 'x0_range' in self.sn_pars:
            xmin, xmax = self.sn_pars['x0_range']
            x0 = np.random.uniform(xmin, xmax, size=self.nsn)
        else:
            x0 = np.full(self.nsn, 1.)
        if 'tmax_margin' in self.sn_pars:
            tmax = np.random.uniform(self.xmin+5, self.xmax-5, size=self.nsn)
        else:
            tmax = np.full(self.nsn, 0.)
        if 'stretch_range' in self.sn_pars:
            stretch = np.random.normal(0., 0.25, size=self.nsn)
        else:
            stretch = np.full(self.nsn, 0.)
        if 'color_pars' in self.sn_pars:
            loc, sig = self.sn_pars['color_pars']
            color = np.random.normal(loc=loc, scale=sig, size=self.nsn)

        sample = pandas.DataFrame({'x0': x0, 'tmax': tmax, 'stretch': stretch, 'color': color})

        return sample

    def gen_data(self, sample):
        """
        """
        # generate the data structure
        npts = self.nsn * self.nbands * self.npts
        d = {
            'x': np.random.uniform(self.xmin, self.xmax, npts),
            'sn': np.repeat(np.arange(self.nsn), self.nbands * self.npts),
            'band': np.tile(np.repeat(np.arange(self.nbands), self.npts), self.nsn),
            'y': np.zeros(npts),
            'yerr': np.zeros(npts),
        }

        for band in self.lcpars:
            d['wl'] = 1500. * d['band'] + 4000.
            dwl = (d['wl'] - self.wl_pivot) / self.wl_pivot
            cl = 1. + dwl * sample.color[d['sn']]
            dt, sigma = self.lcpars[band]
            ph = (d['x']-sample.tmax[d['sn']]-dt) * (1. + sample.stretch[d['sn']])
            d['y'] = sample.x0[d['sn']] * gaussian(ph, sigma) * cl
            d['y_true'] = d['y']
            d['yerr'] = sample.x0[d['sn']] * np.full(npts, self.yerr)
            d['var'] = np.sqrt(self.error_pedestal**2 + self.error_snake**2 * d['y']**2)

        d = pandas.DataFrame(d)
        d['sn_band'] = d.sn * 100 + d.band

        # we need to index the sn_band couples to generate the color scatter
        dp = DataProxy(d, x='x', sn='sn', band='band', sn_band='sn_band',
                       wl='wl', y='y', yerr='yerr', var='var', y_true='y_true')
        dp.make_index('sn_band')

        # add correlated noise
        if self.v_calib is not None:
            L = np.linalg.cholesky(self.v_calib)
            eta = np.random.normal(0., scale=1., size=self.nbands)
            calib_scatter = L @ eta
            d['calib_scatter'] = 1. + calib_scatter[d.band]
            d['y'] *= d.calib_scatter

        if self.color_scatter is not None:
            wl = np.unique(np.vstack((dp.sn_band_index, dp.wl)).T, axis=0)[:,1]
            rwl = (wl - self.wl_pivot) / self.wl_pivot
            sig = np.polynomial.polynomial.Polynomial(self.color_scatter)(rwl)
            kappa = np.random.normal(0., scale=sig)
            d['color_scatter'] = 1. + kappa[dp.sn_band_index]
            d['y'] *= d.color_scatter

        # finally, add some uncorrelated noise
        sig = np.sqrt(d.yerr**2 + self.error_pedestal**2 + self.error_snake**2 * d['y']**2)
        d['noise'] = np.random.normal(scale=sig, size=len(sig))
        d.y += d['noise']

        # final dataproxy, with the final indexes
        dp = DataProxy(d, x='x', sn='sn', band='band', sn_band='sn_band',
                       wl='wl', y='y', yerr='yerr', var='var', y_true='y_true')
        dp.add_field('bads', np.zeros(len(dp.nt)).astype(bool))
        dp.make_index('sn')
        dp.make_index('band')
        dp.make_index('sn_band')

        return dp

    def gen_fine_grid(self, npts_per_lc=100):
        """
        """
        npts = 1 * self.nbands * npts_per_lc
        d = {
            'x': np.tile(np.linspace(self.xmin, self.xmax, npts_per_lc), self.nbands),
            'sn': np.zeros(npts).astype(int),
            'band': np.repeat(np.arange(self.nbands), npts_per_lc).astype(int),
            'y': np.zeros(npts),
            'yerr': np.zeros(npts),
            'var': np.zeros(npts),
            'bads': np.zeros(npts).astype(bool)
        }
        d = pandas.DataFrame(d)
        d['wl'] = 1500. * d['band'] + 4000.
        d['sn_band'] = d.sn * 100 + d.band

        dp = DataProxy(d, x='x', sn='sn', band='band', sn_band='sn_band', wl='wl', y='y', yerr='yerr', var='var')
        dp.make_index('sn')
        dp.make_index('band')
        dp.make_index('sn_band')

        return dp
    
    def generate(self):
        """
        """
        sample = self.gen_sample()
        dp = self.gen_data(sample)
        dp.sample = sample
        dp.band_wl = self.band_wl
        return dp
    def change_noise(self, dp):
        if self.v_calib is not None:
            L = np.linalg.cholesky(self.v_calib)
            eta = np.random.normal(0., scale=1., size=self.nbands)
            calib_scatter = L @ eta
            dp.nt.calib_scatter = 1. + calib_scatter[dp.band]
            dp.y = dp.y_true * dp.nt.calib_scatter
        sig = np.sqrt(dp.yerr**2 + dp.var**2)/5
        noise = np.random.normal(scale=sig, size=len(sig))
        dp.y = dp.y_true + noise



"""Variance models for the simple 1D light curve models


"""

import numpy as np
import scipy.sparse as sparse

class SimplePedestalModel:
    """Just a constant pedestal error added to the data
    """
    def __init__(self, model):
        self.model = model

    def get_struct(self):
        """model specific parameter blocks
        """
        return [("gamma", 1)]

    def init_pars(self, pars):
        pars['gamma'].full[:] = 1.

    def __call__(self, pars, jac=False):
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        N = len(v)
        g = pars['gamma'].full[0]
        var = np.full(N, g**2)

        if not jac:
            return var

        n_free_pars = len(pars.free)
        dvar = np.full(N, 2. * g)
        i = np.arange(N)
        j = pars['gamma'].indexof(np.zeros(N).astype(int))
        idx = j >= 0
        J = sparse.coo_matrix((dvar[idx], (i[idx], j[idx])),
                              shape=(N, n_free_pars))
        return var, J

class SimpleErrorSnake:

    def __init__(self, model):
        self.model = model

    def get_struct(self):
        return [("gamma", 1)]

    def init_pars(self, pars):
        pars['gamma'].full[:] = 0.01

    def __call__(self, pars, jac=False):
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        N = len(v)
        g = pars['gamma'].full[0]
        var = np.full(N, g**2 * v**2)

        if not jac:
            return var

        J = J.tocoo()

        n_free_pars = len(pars.free)
        i = [J.row]
        j = [J.col]
        data = [2. * g**2 * v[J.row] * J.data]
        i.append(np.arange(N))
        j.append(np.full(N, pars['gamma'].indexof(np.zeros(N).astype(int))))
        data.append(2. * g * v**2)

        i = np.hstack(i)
        j = np.hstack(j)
        data = np.hstack(data)

        idx = j >= 0
        J = sparse.coo_matrix((data[idx], (i[idx], j[idx])),
                              shape=(N, n_free_pars))

        return var, J
